from __future__ import annotations

from functools import reduce

from sqlglot import exp, generator, transforms
from sqlglot.dialects.dialect import (
    any_value_to_max_sql,
    date_delta_sql,
    datestrtodate_sql,
    generatedasidentitycolumnconstraint_sql,
    max_or_greatest,
    min_or_least,
    remove_ts_or_ds_to_date,
    rename_func,
    strposition_sql,
    timestrtotime_sql,
    trim_sql,
)
from sqlglot.helper import seq_get
from sqlglot.optimizer.scope import find_in_scope
from sqlglot.parsers.tsql import OPTIONS_THAT_REQUIRE_EQUAL
from sqlglot.time import format_time
from collections import defaultdict

DATE_PART_UNMAPPING = {
    "WEEKISO": "ISO_WEEK",
    "DAYOFWEEK": "WEEKDAY",
    "TIMEZONE_MINUTE": "TZOFFSET",
}

BIT_TYPES = {exp.EQ, exp.NEQ, exp.Is, exp.In, exp.Select, exp.Alias}


def _format_sql(self: TSQLGenerator, expression: exp.NumberToStr | exp.TimeToStr) -> str:
    fmt = expression.args["format"]

    if not isinstance(expression, exp.NumberToStr):
        if fmt.is_string:
            from sqlglot.dialects.tsql import TSQL

            mapped_fmt = format_time(fmt.name, TSQL.INVERSE_TIME_MAPPING)
            fmt_sql = self.sql(exp.Literal.string(mapped_fmt))
        else:
            fmt_sql = self.format_time(expression) or self.sql(fmt)
    else:
        fmt_sql = self.sql(fmt)

    return self.func("FORMAT", expression.this, fmt_sql, expression.args.get("culture"))


def _string_agg_sql(self: TSQLGenerator, expression: exp.GroupConcat) -> str:
    this = expression.this
    distinct = find_in_scope(expression, exp.Distinct)
    if distinct:
        # exp.Distinct can appear below an exp.Order or an exp.GroupConcat expression
        self.unsupported("T-SQL STRING_AGG doesn't support DISTINCT.")
        this = distinct.pop().expressions[0]

    order = ""
    if isinstance(expression.this, exp.Order):
        if expression.this.this:
            this = expression.this.this.pop()
        # Order has a leading space
        order = f" WITHIN GROUP ({self.sql(expression.this)[1:]})"

    separator = expression.args.get("separator") or exp.Literal.string(",")
    return f"STRING_AGG({self.format_args(this, separator)}){order}"


def qualify_derived_table_outputs(expression: exp.Expr) -> exp.Expr:
    """Ensures all (unnamed) output columns are aliased for CTEs and Subqueries."""
    alias = expression.args.get("alias")

    if (
        isinstance(expression, (exp.CTE, exp.Subquery))
        and isinstance(alias, exp.TableAlias)
        and not alias.columns
    ):
        from sqlglot.dialects.tsql import TSQL
        from sqlglot.optimizer.qualify_columns import qualify_outputs

        # We keep track of the unaliased column projection indexes instead of the expressions
        # themselves, because the latter are going to be replaced by new nodes when the aliases
        # are added and hence we won't be able to reach these newly added Alias parents
        query = expression.this
        unaliased_column_indexes = (
            i for i, c in enumerate(query.selects) if isinstance(c, exp.Column) and not c.alias
        )

        qualify_outputs(query, dialect=TSQL())

        # Preserve the quoting information of columns for newly added Alias nodes
        query_selects = query.selects
        for select_index in unaliased_column_indexes:
            alias = query_selects[select_index]
            column = alias.this
            if isinstance(column.this, exp.Identifier):
                alias.args["alias"].set("quoted", column.this.quoted)

    return expression


def _json_extract_sql(
    self: TSQLGenerator, expression: exp.JSONExtract | exp.JSONExtractScalar
) -> str:
    json_query = self.func("JSON_QUERY", expression.this, expression.expression)
    json_value = self.func("JSON_VALUE", expression.this, expression.expression)
    return self.func("ISNULL", json_query, json_value)


def _timestrtotime_sql(self: TSQLGenerator, expression: exp.TimeStrToTime):
    sql = timestrtotime_sql(self, expression)
    if expression.args.get("zone"):
        # If there is a timezone, produce an expression like:
        # CAST('2020-01-01 12:13:14-08:00' AS DATETIMEOFFSET) AT TIME ZONE 'UTC'
        # If you dont have AT TIME ZONE 'UTC', wrapping that expression in another cast back to DATETIME2 just drops the timezone information
        return self.sql(exp.AtTimeZone(this=sql, zone=exp.Literal.string("UTC")))
    return sql


class TSQLGenerator(generator.Generator):
    SELECT_KINDS: tuple[str, ...] = ()
    TRY_SUPPORTED = False
    SUPPORTS_UESCAPE = False
    SUPPORTS_DECODE_CASE = False

    AFTER_HAVING_MODIFIER_TRANSFORMS = generator.AFTER_HAVING_MODIFIER_TRANSFORMS

    LIMIT_IS_TOP = True
    QUERY_HINTS = False
    RETURNING_END = False
    NVL2_SUPPORTED = False
    ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
    LIMIT_FETCH = "FETCH"
    COMPUTED_COLUMN_WITH_TYPE = False
    CTE_RECURSIVE_KEYWORD_REQUIRED = False
    ENSURE_BOOLS = True
    NULL_ORDERING_SUPPORTED: bool | None = None
    SUPPORTS_SINGLE_ARG_CONCAT = False
    TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
    SUPPORTS_SELECT_INTO = True
    JSON_PATH_BRACKETED_KEY_SUPPORTED = False
    SUPPORTS_TO_NUMBER = False
    SET_OP_MODIFIERS = False
    COPY_PARAMS_EQ_REQUIRED = True
    PARSE_JSON_NAME: str | None = None
    EXCEPT_INTERSECT_SUPPORT_ALL_CLAUSE = False
    ALTER_SET_WRAPPED = True
    ALTER_SET_TYPE = ""
    SUPPORTS_ALTER_COLUMN_NULLABILITY = True

    EXPRESSIONS_WITHOUT_NESTED_CTES = {
        exp.Create,
        exp.Delete,
        exp.Insert,
        exp.Intersect,
        exp.Except,
        exp.Merge,
        exp.Select,
        exp.Subquery,
        exp.Union,
        exp.Update,
    }

    SUPPORTED_JSON_PATH_PARTS = {
        exp.JSONPathKey,
        exp.JSONPathRoot,
        exp.JSONPathSubscript,
    }

    TYPE_MAPPING = {
        **{
            k: v
            for k, v in generator.Generator.TYPE_MAPPING.items()
            if k not in (exp.DType.NCHAR, exp.DType.NVARCHAR)
        },
        exp.DType.BOOLEAN: "BIT",
        exp.DType.DATETIME2: "DATETIME2",
        exp.DType.DECIMAL: "NUMERIC",
        exp.DType.DOUBLE: "FLOAT",
        exp.DType.INT: "INTEGER",
        exp.DType.ROWVERSION: "ROWVERSION",
        exp.DType.TEXT: "VARCHAR(MAX)",
        exp.DType.TIMESTAMP: "DATETIME2",
        exp.DType.TIMESTAMPNTZ: "DATETIME2",
        exp.DType.TIMESTAMPTZ: "DATETIMEOFFSET",
        exp.DType.SMALLDATETIME: "SMALLDATETIME",
        exp.DType.UTINYINT: "TINYINT",
        exp.DType.VARIANT: "SQL_VARIANT",
        exp.DType.UUID: "UNIQUEIDENTIFIER",
    }

    TRANSFORMS = {
        **{k: v for k, v in generator.Generator.TRANSFORMS.items() if k != exp.ReturnsProperty},
        exp.AnyValue: any_value_to_max_sql,
        exp.Atan2: rename_func("ATN2"),
        exp.ArrayToString: rename_func("STRING_AGG"),
        exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
        exp.Ceil: rename_func("CEILING"),
        exp.Chr: rename_func("CHAR"),
        exp.DateAdd: date_delta_sql("DATEADD"),
        exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
        exp.CurrentDate: rename_func("GETDATE"),
        exp.CurrentTimestamp: rename_func("GETDATE"),
        exp.CurrentTimestampLTZ: rename_func("SYSDATETIMEOFFSET"),
        exp.DateStrToDate: datestrtodate_sql,
        exp.Day: remove_ts_or_ds_to_date(),
        exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
        exp.GroupConcat: _string_agg_sql,
        exp.If: rename_func("IIF"),
        exp.JSONExtract: _json_extract_sql,
        exp.JSONExtractScalar: _json_extract_sql,
        exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
        exp.Ln: rename_func("LOG"),
        exp.Max: max_or_greatest,
        exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
        exp.Min: min_or_least,
        exp.Month: remove_ts_or_ds_to_date(),
        exp.NumberToStr: _format_sql,
        exp.Repeat: rename_func("REPLICATE"),
        exp.CurrentSchema: rename_func("SCHEMA_NAME"),
        exp.Select: transforms.preprocess(
            [
                transforms.eliminate_distinct_on,
                transforms.eliminate_semi_and_anti_joins,
                transforms.eliminate_qualify,
                transforms.unnest_generate_date_array_using_recursive_cte,
            ]
        ),
        exp.Stddev: rename_func("STDEV"),
        exp.StrPosition: lambda self, e: strposition_sql(
            self, e, func_name="CHARINDEX", supports_position=True
        ),
        exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
        exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
        exp.SHA1Digest: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
        exp.SHA2: lambda self, e: self.func(
            "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
        ),
        exp.TemporaryProperty: lambda self, e: "",
        exp.TimeStrToTime: _timestrtotime_sql,
        exp.TimeToStr: _format_sql,
        exp.TimestampAdd: date_delta_sql("DATEADD"),
        exp.Trim: trim_sql,
        exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
        exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
        exp.TimestampTrunc: lambda self, e: self.func("DATETRUNC", e.unit, e.this),
        exp.Trunc: lambda self, e: self.func(
            "ROUND",
            e.this,
            e.args.get("decimals") or exp.Literal.number(0),
            exp.Literal.number(1),
        ),
        exp.Uuid: lambda *_: "NEWID()",
        exp.Year: remove_ts_or_ds_to_date(),
        exp.DateFromParts: rename_func("DATEFROMPARTS"),
    }

    PROPERTIES_LOCATION = {
        **generator.Generator.PROPERTIES_LOCATION,
        exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    }

    def scope_resolution(self, rhs: str, scope_name: str) -> str:
        return f"{scope_name}::{rhs}"

    def select_sql(self, expression: exp.Select) -> str:
        limit = expression.args.get("limit")
        offset = expression.args.get("offset")

        if isinstance(limit, exp.Fetch) and not offset:
            # Dialects like Oracle can FETCH directly from a row set but
            # T-SQL requires an ORDER BY + OFFSET clause in order to FETCH
            offset = exp.Offset(expression=exp.Literal.number(0))
            expression.set("offset", offset)

        if offset:
            if not expression.args.get("order"):
                # ORDER BY is required in order to use OFFSET in a query, so we use
                # a noop order by, since we don't really care about the order.
                # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
                expression.order_by(exp.select(exp.null()).subquery(), copy=False)

            if isinstance(limit, exp.Limit):
                # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
                # we replace here because otherwise TOP would be generated in select_sql
                limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))

        return super().select_sql(expression)

    def convert_sql(self, expression: exp.Convert) -> str:
        name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
        return self.func(name, expression.this, expression.expression, expression.args.get("style"))

    def queryoption_sql(self, expression: exp.QueryOption) -> str:
        option = self.sql(expression, "this")
        value = self.sql(expression, "expression")
        if value:
            optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
            return f"{option} {optional_equal_sign}{value}"
        return option

    def lateral_op(self, expression: exp.Lateral) -> str:
        cross_apply = expression.args.get("cross_apply")
        if cross_apply is True:
            return "CROSS APPLY"
        if cross_apply is False:
            return "OUTER APPLY"

        # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
        self.unsupported("LATERAL clause is not supported.")
        return "LATERAL"

    def splitpart_sql(self, expression: exp.SplitPart) -> str:
        this = expression.this
        split_count = len(this.name.split("."))
        delimiter = expression.args.get("delimiter")
        part_index = expression.args.get("part_index")

        if (
            not all(isinstance(arg, exp.Literal) for arg in (this, delimiter, part_index))
            or (delimiter and delimiter.name != ".")
            or not part_index
            or split_count > 4
        ):
            self.unsupported(
                "SPLIT_PART can be transpiled to PARSENAME only for '.' delimiter and literal values"
            )
            return ""

        return self.func(
            "PARSENAME", this, exp.Literal.number(split_count + 1 - part_index.to_py())
        )

    def extract_sql(self, expression: exp.Extract) -> str:
        part = expression.this
        name = DATE_PART_UNMAPPING.get(part.name.upper()) or part

        return self.func("DATEPART", name, expression.expression)

    def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
        nano = expression.args.get("nano")
        if nano is not None:
            nano.pop()
            self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")

        if expression.args.get("fractions") is None:
            expression.set("fractions", exp.Literal.number(0))
        if expression.args.get("precision") is None:
            expression.set("precision", exp.Literal.number(0))

        return rename_func("TIMEFROMPARTS")(self, expression)

    def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
        zone = expression.args.get("zone")
        if zone is not None:
            zone.pop()
            self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")

        nano = expression.args.get("nano")
        if nano is not None:
            nano.pop()
            self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")

        if expression.args.get("milli") is None:
            expression.set("milli", exp.Literal.number(0))

        return rename_func("DATETIMEFROMPARTS")(self, expression)

    def setitem_sql(self, expression: exp.SetItem) -> str:
        this = expression.this
        if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
            # T-SQL does not use '=' in SET command, except when the LHS is a variable.
            return f"{self.sql(this.left)} {self.sql(this.right)}"

        return super().setitem_sql(expression)

    def boolean_sql(self, expression: exp.Boolean) -> str:
        if type(expression.parent) in BIT_TYPES or isinstance(
            expression.find_ancestor(exp.Values, exp.Select), exp.Values
        ):
            return "1" if expression.this else "0"

        return "(1 = 1)" if expression.this else "(1 = 0)"

    def is_sql(self, expression: exp.Is) -> str:
        negate = expression.args.get("negate")
        if isinstance(expression.expression, exp.Boolean):
            return self.binary(expression, "<>" if negate else "=")
        return self.binary(expression, "IS NOT" if negate else "IS")

    def createable_sql(self, expression: exp.Create, locations: defaultdict) -> str:
        sql = self.sql(expression, "this")
        properties = expression.args.get("properties")

        start = self._identifier_start
        if (
            not sql.startswith("#")
            and not sql.startswith(f"{start}#")
            and any(
                isinstance(prop, exp.TemporaryProperty)
                for prop in (properties.expressions if properties else [])
            )
        ):
            sql = f"{start}#{sql[len(start) :]}" if sql.startswith(start) else f"#{sql}"

        return sql

    def create_sql(self, expression: exp.Create) -> str:
        kind = expression.kind
        exists = expression.args.get("exists")
        expression.set("exists", None)

        like_property = expression.find(exp.LikeProperty)
        if like_property:
            ctas_expression = like_property.this
        else:
            ctas_expression = expression.expression

        if kind == "VIEW":
            expression.this.set("catalog", None)
            with_ = expression.args.get("with_")
            if ctas_expression and with_:
                # We've already preprocessed the Create expression to bubble up any nested CTEs,
                # but CREATE VIEW actually requires the WITH clause to come after it so we need
                # to amend the AST by moving the CTEs to the CREATE VIEW statement's query.
                ctas_expression.set("with_", with_.pop())
        elif (
            kind == "FUNCTION"
            and isinstance(ctas_expression, exp.Return)
            and isinstance(body := ctas_expression.this.unnest(), exp.Query)
            and (with_ := expression.args.get("with_"))
        ):
            # Similar to the VIEW branch, the table-valued functions require the WITH clause
            # to stay inside the RETURN body, so we move back any CTEs that were bubbled up.
            body.set("with_", with_.pop())

        table = expression.find(exp.Table)

        # Convert CTAS statement to SELECT .. INTO ..
        if kind == "TABLE" and ctas_expression:
            if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
                ctas_expression = ctas_expression.subquery()

            properties = expression.args.get("properties") or exp.Properties()
            is_temp = any(isinstance(p, exp.TemporaryProperty) for p in properties.expressions)

            select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
            select_into.set("into", exp.Into(this=table, temporary=is_temp))

            if like_property:
                select_into.limit(0, copy=False)

            sql = self.sql(select_into)
        else:
            sql = super().create_sql(expression)

        if exists:
            identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
            sql_with_ctes = self.prepend_ctes(expression, sql)
            sql_literal = self.sql(exp.Literal.string(sql_with_ctes))
            if kind == "SCHEMA":
                return f"""IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = {identifier}) EXEC({sql_literal})"""
            elif kind == "TABLE":
                assert table
                where = exp.and_(
                    exp.column("TABLE_NAME").eq(table.name),
                    exp.column("TABLE_SCHEMA").eq(table.db) if table.db else None,
                    exp.column("TABLE_CATALOG").eq(table.catalog) if table.catalog else None,
                )
                return f"""IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE {where}) EXEC({sql_literal})"""
            elif kind == "INDEX":
                index = self.sql(exp.Literal.string(expression.this.text("this")))
                return f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql_literal})"""
        elif expression.args.get("replace"):
            sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)

        return self.prepend_ctes(expression, sql)

    @generator.unsupported_args("unlogged", "expressions")
    def into_sql(self, expression: exp.Into) -> str:
        if expression.args.get("temporary"):
            # If the Into expression has a temporary property, push this down to the Identifier
            table = expression.find(exp.Table)
            if table and isinstance(table.this, exp.Identifier):
                table.this.set("temporary", True)

        return f"{self.seg('INTO')} {self.sql(expression, 'this')}"

    def count_sql(self, expression: exp.Count) -> str:
        func_name = "COUNT_BIG" if expression.args.get("big_int") else "COUNT"
        return rename_func(func_name)(self, expression)

    def datediff_sql(self, expression: exp.DateDiff) -> str:
        func_name = "DATEDIFF_BIG" if expression.args.get("big_int") else "DATEDIFF"
        return date_delta_sql(func_name)(self, expression)

    def offset_sql(self, expression: exp.Offset) -> str:
        return f"{super().offset_sql(expression)} ROWS"

    def version_sql(self, expression: exp.Version) -> str:
        name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
        this = f"FOR {name}"
        expr = expression.expression
        kind = expression.text("kind")
        if kind in ("FROM", "BETWEEN"):
            args = expr.expressions
            sep = "TO" if kind == "FROM" else "AND"
            expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
        else:
            expr_sql = self.sql(expr)

        expr_sql = f" {expr_sql}" if expr_sql else ""
        return f"{this} {kind}{expr_sql}"

    def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
        table = expression.args.get("table")
        table = f"{table} " if table else ""
        return f"RETURNS {table}{self.sql(expression, 'this')}"

    def returning_sql(self, expression: exp.Returning) -> str:
        into = self.sql(expression, "into")
        into = self.seg(f"INTO {into}") if into else ""
        return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"

    def transaction_sql(self, expression: exp.Transaction) -> str:
        this = self.sql(expression, "this")
        this = f" {this}" if this else ""
        mark = self.sql(expression, "mark")
        mark = f" WITH MARK {mark}" if mark else ""
        return f"BEGIN TRANSACTION{this}{mark}"

    def commit_sql(self, expression: exp.Commit) -> str:
        this = self.sql(expression, "this")
        this = f" {this}" if this else ""
        durability = expression.args.get("durability")
        durability = (
            f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
            if durability is not None
            else ""
        )
        return f"COMMIT TRANSACTION{this}{durability}"

    def rollback_sql(self, expression: exp.Rollback) -> str:
        this = self.sql(expression, "this")
        this = f" {this}" if this else ""
        return f"ROLLBACK TRANSACTION{this}"

    def identifier_sql(self, expression: exp.Identifier) -> str:
        identifier = super().identifier_sql(expression)

        if expression.args.get("global_"):
            prefix = "##"
        elif expression.args.get("temporary"):
            prefix = "#"
        else:
            return identifier

        start = self._identifier_start
        if expression.quoted and identifier.startswith(start):
            return f"{start}{prefix}{identifier[len(start) :]}"

        return f"{prefix}{identifier}"

    def constraint_sql(self, expression: exp.Constraint) -> str:
        this = self.sql(expression, "this")
        expressions = self.expressions(expression, flat=True, sep=" ")
        return f"CONSTRAINT {this} {expressions}"

    def length_sql(self, expression: exp.Length) -> str:
        return self._uncast_text(expression, "LEN")

    def right_sql(self, expression: exp.Right) -> str:
        return self._uncast_text(expression, "RIGHT")

    def left_sql(self, expression: exp.Left) -> str:
        return self._uncast_text(expression, "LEFT")

    def _uncast_text(self, expression: exp.Expr, name: str) -> str:
        this = expression.this
        if isinstance(this, exp.Cast) and this.is_type(exp.DType.TEXT):
            this_sql = self.sql(this, "this")
        else:
            this_sql = self.sql(this)
        expression_sql = self.sql(expression, "expression")
        return self.func(name, this_sql, expression_sql if expression_sql else None)

    def partition_sql(self, expression: exp.Partition) -> str:
        return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"

    def alter_sql(self, expression: exp.Alter) -> str:
        action = seq_get(expression.args.get("actions") or [], 0)
        if isinstance(action, exp.AlterRename):
            return f"EXEC sp_rename '{self.sql(expression.this)}', '{action.this.name}'"
        return super().alter_sql(expression)

    def drop_sql(self, expression: exp.Drop) -> str:
        if expression.args["kind"] == "VIEW":
            expression.this.set("catalog", None)
        return super().drop_sql(expression)

    def options_modifier(self, expression: exp.Expr) -> str:
        options = self.expressions(expression, key="options")
        return f" OPTION{self.wrap(options)}" if options else ""

    def dpipe_sql(self, expression: exp.DPipe) -> str:
        return self.sql(reduce(lambda x, y: exp.Add(this=x, expression=y), expression.flatten()))

    def isascii_sql(self, expression: exp.IsAscii) -> str:
        return f"(PATINDEX(CONVERT(VARCHAR(MAX), 0x255b5e002d7f5d25) COLLATE Latin1_General_BIN, {self.sql(expression.this)}) = 0)"

    def columndef_sql(self, expression: exp.ColumnDef, sep: str = " ") -> str:
        this = super().columndef_sql(expression, sep)
        default = self.sql(expression, "default")
        default = f" = {default}" if default else ""
        output = self.sql(expression, "output")
        output = f" {output}" if output else ""
        return f"{this}{default}{output}"

    def coalesce_sql(self, expression: exp.Coalesce) -> str:
        func_name = "ISNULL" if expression.args.get("is_null") else "COALESCE"
        return rename_func(func_name)(self, expression)

    def storedprocedure_sql(self, expression: exp.StoredProcedure) -> str:
        this = self.sql(expression, "this")
        expressions = self.expressions(expression)
        expressions = (
            self.wrap(expressions) if expression.args.get("wrapped") else f" {expressions}"
        )
        return f"{this}{expressions}" if expressions.strip() != "" else this

    def ifblock_sql(self, expression: exp.IfBlock) -> str:
        this = self.sql(expression, "this")
        true = self.sql(expression, "true")
        true = f" {true}" if true else " "
        false = self.sql(expression, "false")
        false = f"; ELSE BEGIN {false}" if false else ""
        return f"IF {this} BEGIN{true}{false}"

    def whileblock_sql(self, expression: exp.WhileBlock) -> str:
        this = self.sql(expression, "this")
        body = self.sql(expression, "body")
        body = f" {body}" if body else " "
        return f"WHILE {this} BEGIN{body}"

    def execute_sql(self, expression: exp.Execute) -> str:
        this = self.sql(expression, "this")
        expressions = self.expressions(expression)
        expressions = f" {expressions}" if expressions else ""
        return_status = self.sql(expression, "return_status")
        return_status = f"{return_status} = " if return_status else ""
        return f"EXECUTE {return_status}{this}{expressions}"

    def executesql_sql(self, expression: exp.ExecuteSql) -> str:
        return self.execute_sql(expression)



import typing as t

from sqlglot import exp
from sqlglot.helper import subclasses
from builtins import type as Type

ExprMetadataType = dict[Type[exp.Expr], dict[str, t.Any]]

TIMESTAMP_EXPRESSIONS = {
    exp.CurrentTimestamp,
    exp.StrToTime,
    exp.TimeStrToTime,
    exp.TimestampAdd,
    exp.TimestampSub,
    exp.UnixToTime,
}

EXPRESSION_METADATA: ExprMetadataType = {
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_binary(e)}
        for expr_type in subclasses(exp.__name__, exp.Binary)
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_unary(e)}
        for expr_type in subclasses(
            exp.__name__, (exp.Unary, exp.Alias, exp.IgnoreNulls, exp.RespectNulls)
        )
    },
    **{
        expr_type: {"returns": exp.DType.BIGINT}
        for expr_type in {
            exp.ApproxDistinct,
            exp.ArraySize,
            exp.CountIf,
            exp.DenseRank,
            exp.Int64,
            exp.Ntile,
            exp.Rank,
            exp.RowNumber,
            exp.UnixSeconds,
            exp.UnixMicros,
            exp.UnixMillis,
        }
    },
    **{
        expr_type: {"returns": exp.DType.BINARY}
        for expr_type in {
            exp.FromBase32,
            exp.FromBase64,
        }
    },
    **{
        expr_type: {"returns": exp.DType.BOOLEAN}
        for expr_type in {
            exp.All,
            exp.Any,
            exp.Between,
            exp.Boolean,
            exp.Contains,
            exp.EndsWith,
            exp.Exists,
            exp.In,
            exp.IsInf,
            exp.IsNan,
            exp.LogicalAnd,
            exp.LogicalOr,
            exp.StartsWith,
        }
    },
    **{
        expr_type: {"returns": exp.DType.DATE}
        for expr_type in {
            exp.CurrentDate,
            exp.Date,
            exp.DateFromParts,
            exp.DateStrToDate,
            exp.DiToDate,
            exp.LastDay,
            exp.StrToDate,
            exp.TimeStrToDate,
            exp.TsOrDsToDate,
        }
    },
    **{
        expr_type: {"returns": exp.DType.DATETIME}
        for expr_type in {
            exp.CurrentDatetime,
            exp.Datetime,
            exp.DatetimeAdd,
            exp.DatetimeSub,
        }
    },
    **{
        expr_type: {"returns": exp.DType.DOUBLE}
        for expr_type in {
            exp.Asin,
            exp.Asinh,
            exp.Acos,
            exp.CovarPop,
            exp.CovarSamp,
            exp.Acosh,
            exp.ApproxQuantile,
            exp.Atan,
            exp.Atanh,
            exp.Avg,
            exp.Cbrt,
            exp.Cos,
            exp.Cosh,
            exp.Cot,
            exp.Degrees,
            exp.Exp,
            exp.Kurtosis,
            exp.Ln,
            exp.Log,
            exp.Pi,
            exp.Pow,
            exp.PercentileCont,
            exp.Quantile,
            exp.Radians,
            exp.Round,
            exp.SafeDivide,
            exp.Sin,
            exp.Sinh,
            exp.Sqrt,
            exp.Stddev,
            exp.StddevPop,
            exp.StddevSamp,
            exp.Rand,
            exp.Tan,
            exp.Tanh,
            exp.ToDouble,
            exp.CumeDist,
            exp.PercentRank,
            exp.Variance,
            exp.VariancePop,
            exp.Skewness,
        }
    },
    **{
        expr_type: {"returns": exp.DType.INT}
        for expr_type in {
            exp.Ascii,
            exp.BitLength,
            exp.Ceil,
            exp.DatetimeDiff,
            exp.DayOfMonth,
            exp.DayOfWeek,
            exp.DayOfYear,
            exp.Floor,
            exp.Getbit,
            exp.Hour,
            exp.TimestampDiff,
            exp.TimeDiff,
            exp.Unicode,
            exp.DateToDi,
            exp.Levenshtein,
            exp.Length,
            exp.Sign,
            exp.StrPosition,
            exp.TsOrDiToDi,
            exp.Quarter,
            exp.UnixDate,
        }
    },
    **{
        expr_type: {"returns": exp.DType.INTERVAL}
        for expr_type in {
            exp.Interval,
            exp.JustifyDays,
            exp.JustifyHours,
            exp.JustifyInterval,
            exp.MakeInterval,
        }
    },
    **{
        expr_type: {"returns": exp.DType.JSON}
        for expr_type in {
            exp.ParseJSON,
        }
    },
    **{
        expr_type: {"returns": exp.DType.TIME}
        for expr_type in {
            exp.CurrentTime,
            exp.Localtime,
            exp.Time,
            exp.TimeAdd,
            exp.TimeSub,
        }
    },
    **{
        expr_type: {"returns": exp.DType.TIMESTAMPLTZ}
        for expr_type in {
            exp.TimestampLtzFromParts,
        }
    },
    **{
        expr_type: {"returns": exp.DType.TIMESTAMPTZ}
        for expr_type in {
            exp.CurrentTimestampLTZ,
            exp.TimestampTzFromParts,
        }
    },
    **{expr_type: {"returns": exp.DType.TIMESTAMP} for expr_type in TIMESTAMP_EXPRESSIONS},
    **{
        expr_type: {"returns": exp.DType.TINYINT}
        for expr_type in {
            exp.Day,
            exp.DayOfWeekIso,
            exp.Month,
            exp.Week,
            exp.WeekOfYear,
            exp.Year,
            exp.YearOfWeek,
            exp.YearOfWeekIso,
        }
    },
    **{
        expr_type: {"returns": exp.DType.VARCHAR}
        for expr_type in {
            exp.ArrayToString,
            exp.Concat,
            exp.ConcatWs,
            exp.Chr,
            exp.CurrentCatalog,
            exp.CurrentRole,
            exp.CurrentSchema,
            exp.CurrentVersion,
            exp.CurrentUser,
            exp.Dayname,
            exp.DateToDateStr,
            exp.DPipe,
            exp.GroupConcat,
            exp.Initcap,
            exp.Lower,
            exp.MD5,
            exp.Monthname,
            exp.RawString,
            exp.Repeat,
            exp.SHA,
            exp.SHA2,
            exp.SessionUser,
            exp.Space,
            exp.String,
            exp.Substring,
            exp.TimeToStr,
            exp.TimeToTimeStr,
            exp.Trim,
            exp.ToBase32,
            exp.ToBase64,
            exp.Translate,
            exp.TsOrDsToDateStr,
            exp.Typeof,
            exp.UnixToStr,
            exp.UnixToTimeStr,
            exp.Upper,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_by_args(e, "this")}
        for expr_type in {
            exp.Abs,
            exp.AnyValue,
            exp.ArrayConcatAgg,
            exp.ArrayReverse,
            exp.ArraySlice,
            exp.Filter,
            exp.FirstValue,
            exp.HavingMax,
            exp.LastValue,
            exp.Limit,
            exp.NthValue,
            exp.Order,
            exp.SortArray,
            exp.Window,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_by_args(e, "this", "expressions")}
        for expr_type in {
            exp.ArrayConcat,
            exp.Coalesce,
            exp.Greatest,
            exp.Least,
            exp.Max,
            exp.Min,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_by_array_element(e)}
        for expr_type in {
            exp.ArrayFirst,
            exp.ArrayLast,
        }
    },
    exp.Anonymous: {"annotator": lambda self, e: self._set_type(e, self.schema.get_udf_type(e))},
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_timeunit(e)}
        for expr_type in {
            exp.DateAdd,
            exp.DateSub,
            exp.DateTrunc,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._set_type(e, e.args["to"])}
        for expr_type in {
            exp.Cast,
            exp.TryCast,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_map(e)}
        for expr_type in {
            exp.Map,
            exp.VarMap,
        }
    },
    exp.Array: {"annotator": lambda self, e: self._annotate_by_args(e, "expressions", array=True)},
    exp.ArrayAgg: {"annotator": lambda self, e: self._annotate_by_args(e, "this", array=True)},
    exp.Bracket: {"annotator": lambda self, e: self._annotate_bracket(e)},
    exp.Case: {
        "annotator": lambda self, e: self._annotate_by_args(
            e, *[if_expr.args["true"] for if_expr in e.args["ifs"]], "default"
        )
    },
    exp.Count: {
        "annotator": lambda self, e: self._set_type(
            e, exp.DType.BIGINT if e.args.get("big_int") else exp.DType.INT
        )
    },
    exp.DateDiff: {
        "annotator": lambda self, e: self._set_type(
            e, exp.DType.BIGINT if e.args.get("big_int") else exp.DType.INT
        )
    },
    exp.DataType: {"annotator": lambda _, e: e},
    exp.Div: {"annotator": lambda self, e: self._annotate_div(e)},
    exp.Distinct: {"annotator": lambda self, e: self._annotate_by_args(e, "expressions")},
    exp.Dot: {"annotator": lambda self, e: self._annotate_dot(e)},
    exp.Explode: {"annotator": lambda self, e: self._annotate_explode(e)},
    exp.Extract: {"annotator": lambda self, e: self._annotate_extract(e)},
    exp.HexString: {
        "annotator": lambda self, e: self._set_type(
            e,
            exp.DType.BIGINT if e.args.get("is_integer") else exp.DType.BINARY,
        )
    },
    exp.GenerateSeries: {
        "annotator": lambda self, e: self._annotate_by_args(e, "start", "end", "step", array=True)
    },
    exp.GenerateDateArray: {
        "annotator": lambda self, e: self._set_type(e, exp.DataType.from_str("ARRAY<DATE>"))
    },
    exp.GenerateTimestampArray: {
        "annotator": lambda self, e: self._set_type(e, exp.DataType.from_str("ARRAY<TIMESTAMP>"))
    },
    exp.If: {"annotator": lambda self, e: self._annotate_by_args(e, "true", "false")},
    exp.Lag: {"annotator": lambda self, e: self._annotate_by_args(e, "this", "default")},
    exp.Lead: {"annotator": lambda self, e: self._annotate_by_args(e, "this", "default")},
    exp.Literal: {"annotator": lambda self, e: self._annotate_literal(e)},
    exp.Null: {"returns": exp.DType.NULL},
    exp.Nullif: {"annotator": lambda self, e: self._annotate_by_args(e, "this", "expression")},
    exp.PropertyEQ: {"annotator": lambda self, e: self._annotate_by_args(e, "expression")},
    exp.Struct: {"annotator": lambda self, e: self._annotate_struct(e)},
    exp.Sum: {
        "annotator": lambda self, e: self._annotate_by_args(e, "this", "expressions", promote=True)
    },
    exp.Timestamp: {
        "annotator": lambda self, e: self._set_type(
            e,
            exp.DType.TIMESTAMPTZ if e.args.get("with_tz") else exp.DType.TIMESTAMP,
        )
    },
    exp.ToMap: {"annotator": lambda self, e: self._annotate_to_map(e)},
    exp.Unnest: {"annotator": lambda self, e: self._annotate_unnest(e)},
    exp.WithinGroup: {"annotator": lambda self, e: self._annotate_within_group(e)},
    exp.Subquery: {"annotator": lambda self, e: self._annotate_subquery(e)},
}



from __future__ import annotations

import itertools
import typing as t

from sqlglot import exp
from sqlglot.dialects.dialect import Dialect
from sqlglot.errors import OptimizeError
from sqlglot.helper import seq_get, SingleValuedMapping
from sqlglot.optimizer.scope import Scope

if t.TYPE_CHECKING:
    from sqlglot.schema import Schema
    from collections.abc import Sequence, Mapping


class Resolver:
    """
    Helper for resolving columns.

    This is a class so we can lazily load some things and easily share them across functions.
    """

    def __init__(self, scope: Scope, schema: Schema, infer_schema: bool = True) -> None:
        self.scope: Scope = scope
        self.schema: Schema = schema
        self.dialect: Dialect = schema.dialect or Dialect()
        self._source_columns: dict[str, Sequence[str]] | None = None
        self._unambiguous_columns: Mapping[str, str] | None = None
        self._all_columns: set[str] | None = None
        self._infer_schema: bool = infer_schema
        self._get_source_columns_cache: dict[tuple[str, bool], Sequence[str]] = {}
        self._column_type_from_scope_cache: dict[tuple[int, str], exp.DataType | None] = {}

    def get_table(self, column: str | exp.Column) -> exp.Identifier | None:
        """
        Get the table for a column name.

        Args:
            column: The column expression (or column name) to find the table for.
        Returns:
            The table name if it can be found/inferred.
        """
        column_name = column if isinstance(column, str) else column.name

        table_name = self._get_table_name_from_sources(column_name)

        if not table_name and isinstance(column, exp.Column):
            # Fall-back case: If we couldn't find the `table_name` from ALL of the sources,
            # attempt to disambiguate the column based on other characteristics e.g if this column is in a join condition,
            # we may be able to disambiguate based on the source order.
            if join_context := self._get_column_join_context(column):
                # In this case, the return value will be the join that _may_ be able to disambiguate the column
                # and we can use the source columns available at that join to get the table name
                # catch OptimizeError if column is still ambiguous and try to resolve with schema inference below
                try:
                    table_name = self._get_table_name_from_sources(
                        column_name, self._get_available_source_columns(join_context)
                    )
                except OptimizeError:
                    pass

        if not table_name and self._infer_schema:
            sources_without_schema = tuple(
                source
                for source, columns in self._get_all_source_columns().items()
                if not columns or "*" in columns
            )
            if len(sources_without_schema) == 1:
                table_name = sources_without_schema[0]

        if table_name not in self.scope.selected_sources:
            return exp.to_identifier(table_name)

        node: exp.Expr = self.scope.selected_sources[table_name][0]

        if isinstance(node, exp.Query):
            while node and node.alias != table_name and node.parent:
                node = node.parent

        node_alias = node.args.get("alias")
        if node_alias:
            return exp.to_identifier(node_alias.this)

        return exp.to_identifier(table_name)

    @property
    def all_columns(self) -> set[str]:
        """All available columns of all sources in this scope"""
        if self._all_columns is None:
            self._all_columns = {
                column for columns in self._get_all_source_columns().values() for column in columns
            }
        return self._all_columns

    def get_source_columns_from_set_op(self, expression: exp.Expr) -> list[str]:
        if isinstance(expression, exp.Select):
            return expression.named_selects
        if isinstance(expression, exp.Subquery) and isinstance(expression.this, exp.SetOperation):
            # Different types of SET modifiers can be chained together if they're explicitly grouped by nesting
            return self.get_source_columns_from_set_op(expression.this)
        if not isinstance(expression, exp.SetOperation):
            raise OptimizeError(f"Unknown set operation: {expression}")

        set_op = expression

        # BigQuery specific set operations modifiers, e.g INNER UNION ALL BY NAME
        on_column_list = set_op.args.get("on")

        if on_column_list:
            # The resulting columns are the columns in the ON clause:
            # {INNER | LEFT | FULL} UNION ALL BY NAME ON (col1, col2, ...)
            columns = [col.name for col in on_column_list]
        elif set_op.side or set_op.kind:
            side = set_op.side
            kind = set_op.kind

            # Visit the children UNIONs (if any) in a post-order traversal
            left = self.get_source_columns_from_set_op(set_op.this)
            right = self.get_source_columns_from_set_op(set_op.expression)

            # We use dict.fromkeys to deduplicate keys and maintain insertion order
            if side == "LEFT":
                columns = left
            elif side == "FULL":
                columns = list(dict.fromkeys(left + right))
            elif kind == "INNER":
                columns = list(dict.fromkeys(left).keys() & dict.fromkeys(right).keys())
        else:
            columns = set_op.named_selects

        return columns

    def get_source_columns(self, name: str, only_visible: bool = False) -> Sequence[str]:
        """Resolve the source columns for a given source `name`."""
        cache_key = (name, only_visible)
        if cache_key not in self._get_source_columns_cache:
            if name not in self.scope.sources:
                raise OptimizeError(f"Unknown table: {name}")

            source = self.scope.sources[name]

            # A pivoted CTE reference is stored as an exp.Table in the scope sources (see
            # _traverse_tables in scope.py), but the underlying CTE Scope still holds the
            # column information we need to resolve pre-pivot columns.
            if (
                isinstance(source, exp.Table)
                and not source.db
                and source.args.get("pivots")
                and source.name in self.scope.cte_sources
            ):
                source = self.scope.cte_sources[source.name]

            if isinstance(source, exp.Table):
                columns = self.schema.column_names(source, only_visible)
            elif isinstance(source, Scope) and isinstance(
                source_expr := source.expression, (exp.Values, exp.Unnest, exp.Lateral)
            ):
                columns = source_expr.named_selects

                # in bigquery, unnest structs are automatically scoped as tables, so you can directly select
                # a struct field in a query. This handles the case where the unnest is statically defined.
                if self.dialect.UNNEST_COLUMN_ONLY and isinstance(source_expr, exp.Unnest):
                    if not source_expr.type or source_expr.type.is_type(exp.DType.UNKNOWN):
                        unnest_expr = seq_get(source_expr.expressions, 0)
                        if isinstance(unnest_expr, exp.Column) and self.scope.parent:
                            col_type = self._get_unnest_column_type(unnest_expr, self.scope.parent)
                            if col_type and col_type.is_type(exp.DType.ARRAY):
                                element_types = col_type.expressions
                                if element_types:
                                    source_expr.type = element_types[0].copy()
                            elif col_type:
                                source_expr.type = col_type.copy()

                    columns.extend(self._struct_field_names(source_expr.type))
                elif isinstance(source_expr, exp.Lateral) and isinstance(
                    source_expr.this, exp.Explode
                ):
                    explode_col = source_expr.this.this

                    # If the column is unqualified at this point, it couldn't be resolved when
                    # this scope's children were qualified; disambiguating it here would require
                    # enumerating this very source's columns, i.e recurse without bound
                    if isinstance(explode_col, exp.Column) and explode_col.table and source.parent:
                        col_type = self._get_unnest_column_type(explode_col, source.parent)
                        columns.extend(self._struct_field_names(col_type))
                elif isinstance(source_expr, exp.Lateral) and isinstance(
                    source_expr.this, exp.Query
                ):
                    columns = source_expr.this.named_selects
            elif isinstance(source, Scope) and isinstance(source.expression, exp.SetOperation):
                columns = self.get_source_columns_from_set_op(source.expression)
            else:
                selectable = source.expression.assert_is(exp.Selectable)
                select = seq_get(selectable.selects, 0)

                if isinstance(select, exp.QueryTransform):
                    # https://spark.apache.org/docs/3.5.1/sql-ref-syntax-qry-select-transform.html
                    schema = select.args.get("schema")
                    columns = [c.name for c in schema.expressions] if schema else ["key", "value"]
                else:
                    columns = selectable.named_selects

            node, _ = self.scope.selected_sources.get(name) or (None, None)
            if isinstance(node, Scope):
                column_aliases = node.expression.alias_column_names
            elif isinstance(node, exp.Expr):
                column_aliases = node.alias_column_names
            else:
                column_aliases = []

            if column_aliases:
                # If the source's columns are aliased, their aliases shadow the corresponding column names.
                # This can be expensive if there are lots of columns, so only do this if column_aliases exist.
                columns = [
                    alias or name
                    for (name, alias) in itertools.zip_longest(columns, column_aliases)
                ]

            self._get_source_columns_cache[cache_key] = columns

        return self._get_source_columns_cache[cache_key]

    def _get_all_source_columns(self) -> dict[str, Sequence[str]]:
        if self._source_columns is None:
            self._source_columns = {
                source_name: self.get_source_columns(source_name)
                for source_name, source in itertools.chain(
                    self.scope.selected_sources.items(), self.scope.lateral_sources.items()
                )
            }
        return self._source_columns

    def _get_table_name_from_sources(
        self, column_name: str, source_columns: dict[str, Sequence[str]] | None = None
    ) -> str | None:
        if not source_columns:
            # If not supplied, get all sources to calculate unambiguous columns
            if self._unambiguous_columns is None:
                self._unambiguous_columns = self._get_unambiguous_columns(
                    self._get_all_source_columns()
                )

            unambiguous_columns = self._unambiguous_columns
        else:
            unambiguous_columns = self._get_unambiguous_columns(source_columns)

        return unambiguous_columns.get(column_name)

    def _get_column_join_context(self, column: exp.Column) -> exp.Join | None:
        """
        Check if a column participating in a join can be qualified based on the source order.
        """
        args = self.scope.expression.args
        joins = args.get("joins")

        if not joins or args.get("laterals") or args.get("pivots"):
            # Feature gap: We currently don't try to disambiguate columns if other sources
            # (e.g laterals, pivots) exist alongside joins
            return None

        join_ancestor = column.find_ancestor(exp.Join, exp.Select)

        if (
            isinstance(join_ancestor, exp.Join)
            and join_ancestor.alias_or_name in self.scope.selected_sources
        ):
            # Ensure that the found ancestor is a join that contains an actual source,
            # e.g in Clickhouse `b` is an array expression in `a ARRAY JOIN b`
            return join_ancestor

        return None

    def _get_available_source_columns(self, join_ancestor: exp.Join) -> dict[str, Sequence[str]]:
        """
        Get the source columns that are available at the point where a column is referenced.

        For columns in JOIN conditions, this only includes tables that have been joined
        up to that point. Example:

        ```
        SELECT * FROM t_1 INNER JOIN ... INNER JOIN t_n ON t_1.a = c INNER JOIN t_n+1 ON ...
        ```                                                        ^
                                                                   |
                                +----------------------------------+
                                |
                                ⌄
        The unqualified column `c` is not ambiguous if no other sources up until that
        join i.e t_1, ..., t_n, contain a column named `c`.

        """
        args = self.scope.expression.args

        # Collect tables in order: FROM clause tables + joined tables up to current join
        from_name = args["from_"].alias_or_name
        available_sources = {from_name: self.get_source_columns(from_name)}

        for join in args["joins"][: t.cast(int, join_ancestor.index) + 1]:
            available_sources[join.alias_or_name] = self.get_source_columns(join.alias_or_name)

        return available_sources

    def _get_unambiguous_columns(
        self, source_columns: dict[str, Sequence[str]]
    ) -> Mapping[str, str]:
        """
        Find all the unambiguous columns in sources.

        Args:
            source_columns: Mapping of names to source columns.

        Returns:
            Mapping of column name to source name.
        """
        if not source_columns:
            return {}

        source_columns_pairs = list(source_columns.items())

        first_table, first_columns = source_columns_pairs[0]

        if len(source_columns_pairs) == 1:
            # Performance optimization - avoid copying first_columns if there is only one table.
            return SingleValuedMapping(first_columns, first_table)

        # For BigQuery UNNEST_COLUMN_ONLY, build a mapping of original UNNEST aliases
        # from alias.columns[0] to their source names. This is used to resolve shadowing
        # where an UNNEST alias shadows a column name from another table.
        unnest_original_aliases: dict[str, str] = {}
        if self.dialect.UNNEST_COLUMN_ONLY:
            unnest_original_aliases = {
                alias_arg.columns[0].name: source_name
                for source_name, source in self.scope.sources.items()
                if (
                    isinstance(source.expression, exp.Unnest)
                    and (alias_arg := source.expression.args.get("alias"))
                    and alias_arg.columns
                )
            }

        unambiguous_columns = {col: first_table for col in first_columns}
        all_columns = set(unambiguous_columns)

        for table, columns in source_columns_pairs[1:]:
            unique = set(columns)
            ambiguous = all_columns.intersection(unique)
            all_columns.update(columns)

            for column in ambiguous:
                if column in unnest_original_aliases:
                    unambiguous_columns[column] = unnest_original_aliases[column]
                    continue

                unambiguous_columns.pop(column, None)
            for column in unique.difference(ambiguous):
                unambiguous_columns[column] = table

        return unambiguous_columns

    def _struct_field_names(self, col_type: exp.DataType | None) -> list[str]:
        if col_type and col_type.is_type(exp.DType.ARRAY):
            col_type = seq_get(col_type.expressions, 0)

        return (
            [k.name for k in col_type.expressions]
            if col_type and col_type.is_type(exp.DType.STRUCT)
            else []
        )

    def _get_unnest_column_type(self, column: exp.Column, scope: Scope) -> exp.DataType | None:
        """
        Get the type of a column being unnested/exploded, tracing through CTEs/subqueries to find the base table.

        Args:
            column: The column expression being unnested/exploded.
            scope: The scope to resolve the column in.

        Returns:
            The DataType of the column, or None if not found.
        """
        # if column is qualified, use that table, otherwise disambiguate using the resolver
        if column.table:
            table_name = column.table
        else:
            # use the parent scope's resolver to disambiguate the column
            parent_resolver = Resolver(scope, self.schema, self._infer_schema)
            table_identifier = parent_resolver.get_table(column)
            if not table_identifier:
                return None
            table_name = table_identifier.name

        source = scope.sources.get(table_name)
        return self._get_column_type_from_scope(source, column) if source else None

    def _get_column_type_from_scope(
        self, source: Scope | exp.Table, column: exp.Column
    ) -> exp.DataType | None:
        """
        Get a column's type by tracing through scopes/tables to find the base table.

        Args:
            source: The source to search - can be a Scope (to iterate its sources) or a Table.
            column: The column to find the type for.

        Returns:
            The DataType of the column, or None if not found.
        """
        # A single source can be reachable through many paths of a scope DAG (e.g. a CTE
        # referenced by several other CTEs). The schema and scope are immutable during
        # qualification, so the type of `column` under `source` depends only on
        # `(source, column name)`; memoize it to walk each source once.
        cache_key = (id(source), column.name)
        if cache_key in self._column_type_from_scope_cache:
            return self._column_type_from_scope_cache[cache_key]

        # None is a valid result if DataType could not be determined!
        result: exp.DataType | None = None
        if isinstance(source, exp.Table):
            # base table - get the column type from schema
            col_type = self.schema.get_column_type(source, column)
            if col_type and not col_type.is_type(exp.DType.UNKNOWN):
                result = col_type
        elif isinstance(source, Scope):
            # iterate over all sources in the scope
            for nested_source in source.sources.values():
                nested_type = self._get_column_type_from_scope(nested_source, column)
                if nested_type and not nested_type.is_type(exp.DType.UNKNOWN):
                    result = nested_type
                    break

        self._column_type_from_scope_cache[cache_key] = result
        return result



import collections
import itertools
import math

from sqlglot import exp, planner, tokens
from sqlglot.dialects.dialect import Dialect
from sqlglot.errors import ExecuteError
from sqlglot.executor.context import Context
from sqlglot.executor.env import ENV
from sqlglot.executor.table import RowReader, Table
from sqlglot.generators.python import PythonGenerator


class PythonExecutor:
    def __init__(self, env=None, tables=None):
        self.generator = Python().generator(identify=True, comments=False)
        self.env = {**ENV, **(env or {})}
        self.tables = tables or {}

    def execute(self, plan):
        finished = set()
        queue = set(plan.leaves)
        contexts = {}

        while queue:
            node = queue.pop()
            try:
                context = self.context(
                    {
                        name: table
                        for dep in node.dependencies
                        for name, table in contexts[dep].tables.items()
                    }
                )

                if isinstance(node, planner.Scan):
                    contexts[node] = self.scan(node, context)
                elif isinstance(node, planner.Aggregate):
                    contexts[node] = self.aggregate(node, context)
                elif isinstance(node, planner.Join):
                    contexts[node] = self.join(node, context)
                elif isinstance(node, planner.Sort):
                    contexts[node] = self.sort(node, context)
                elif isinstance(node, planner.SetOperation):
                    contexts[node] = self.set_operation(node, context)
                else:
                    raise NotImplementedError

                finished.add(node)

                for dep in node.dependents:
                    if all(d in contexts for d in dep.dependencies):
                        queue.add(dep)

                for dep in node.dependencies:
                    if all(d in finished for d in dep.dependents):
                        contexts.pop(dep)
            except Exception as e:
                raise ExecuteError(f"Step '{node.id}' failed: {e}") from e

        root = plan.root
        return contexts[root].tables[root.name]

    def generate(self, expression):
        """Convert a SQL expression into literal Python code and compile it into bytecode."""
        if not expression:
            return None

        sql = self.generator.generate(expression)
        return compile(sql, sql, "eval", optimize=2)

    def generate_tuple(self, expressions):
        """Convert an array of SQL expressions into tuple of Python byte code."""
        if not expressions:
            return tuple()
        return tuple(self.generate(expression) for expression in expressions)

    def context(self, tables):
        return Context(tables, env=self.env)

    def table(self, expressions):
        return Table(
            expression.alias_or_name if isinstance(expression, exp.Expr) else expression
            for expression in expressions
        )

    def scan(self, step, context):
        source = step.source

        if source and isinstance(source, exp.Expr):
            source = source.name or source.alias

        if source is None:
            context, table_iter = self.static()
        elif source in context:
            if not step.projections and not step.condition:
                return self.context({step.name: context.tables[source]})
            table_iter = context.table_iter(source)
        else:
            context, table_iter = self.scan_table(step)

        return self.context({step.name: self._project_and_filter(context, step, table_iter)})

    def _project_and_filter(self, context, step, table_iter):
        sink = self.table(step.projections if step.projections else context.columns)
        condition = self.generate(step.condition)
        projections = self.generate_tuple(step.projections)

        for reader in table_iter:
            if len(sink) >= step.limit:
                break

            if condition and not context.eval(condition):
                continue

            if projections:
                sink.append(context.eval_tuple(projections))
            else:
                sink.append(reader.row)

        return sink

    def static(self):
        return self.context({}), [RowReader(())]

    def scan_table(self, step):
        table = self.tables.find(step.source)
        context = self.context({step.source.alias_or_name: table})
        return context, iter(table)

    def join(self, step, context):
        source = step.source_name

        source_table = context.tables[source]
        source_context = self.context({source: source_table})
        column_ranges = {source: range(0, len(source_table.columns))}

        for name, join in step.joins.items():
            table = context.tables[name]
            start = max(r.stop for r in column_ranges.values())
            column_ranges[name] = range(start, len(table.columns) + start)
            join_context = self.context({name: table})
            condition = self.generate(join["condition"])
            condition_context = (
                self.context(
                    {
                        name: Table(
                            source_context.columns + join_context.columns,
                            column_range=column_range,
                        )
                        for name, column_range in column_ranges.items()
                    }
                )
                if condition
                else None
            )

            if join.get("source_key"):
                table = self.hash_join(
                    join, source_context, join_context, condition, condition_context
                )
            else:
                table = self.nested_loop_join(
                    join, source_context, join_context, condition, condition_context
                )

            source_context = self.context(
                {
                    name: Table(table.columns, table.rows, column_range)
                    for name, column_range in column_ranges.items()
                }
            )
        if not step.condition and not step.projections:
            return source_context

        sink = self._project_and_filter(
            source_context,
            step,
            (reader for reader, _ in iter(source_context)),
        )

        if step.projections:
            return self.context({step.name: sink})
        else:
            return self.context(
                {
                    name: Table(table.columns, sink.rows, table.column_range)
                    for name, table in source_context.tables.items()
                }
            )

    @staticmethod
    def _join_matches(row, condition, condition_context):
        if not condition:
            return True

        condition_context.set_row(row)
        return condition_context.eval(condition) is True

    def nested_loop_join(self, join, source_context, join_context, condition, condition_context):
        table = Table(source_context.columns + join_context.columns)
        source_rows = source_context.table.rows
        join_rows = join_context.table.rows
        matched_source = set()
        matched_join = set()

        for source_index, source_row in enumerate(source_rows):
            for join_index, join_row in enumerate(join_rows):
                row = source_row + join_row
                if self._join_matches(row, condition, condition_context):
                    table.append(row)
                    matched_source.add(source_index)
                    matched_join.add(join_index)

        self._append_unmatched_join_rows(
            table, join, source_rows, join_rows, matched_source, matched_join
        )

        return table

    def hash_join(self, join, source_context, join_context, condition, condition_context):
        source_key = self.generate_tuple(join["source_key"])
        join_key = self.generate_tuple(join["join_key"])
        results = collections.defaultdict(lambda: ([], []))

        for index, (reader, ctx) in enumerate(source_context):
            key = ctx.eval_tuple(source_key)
            if all(value is not None for value in key):
                results[key][0].append((index, reader.row))
        for index, (reader, ctx) in enumerate(join_context):
            key = ctx.eval_tuple(join_key)
            if all(value is not None for value in key):
                results[key][1].append((index, reader.row))

        table = Table(source_context.columns + join_context.columns)
        matched_source = set()
        matched_join = set()

        for source_group, join_group in results.values():
            for (source_index, source_row), (join_index, join_row) in itertools.product(
                source_group, join_group
            ):
                row = source_row + join_row
                if self._join_matches(row, condition, condition_context):
                    table.append(row)
                    matched_source.add(source_index)
                    matched_join.add(join_index)

        self._append_unmatched_join_rows(
            table,
            join,
            source_context.table.rows,
            join_context.table.rows,
            matched_source,
            matched_join,
        )

        return table

    @staticmethod
    def _append_unmatched_join_rows(
        table, join, source_rows, join_rows, matched_source, matched_join
    ):
        side = join.get("side")
        if side in ("LEFT", "FULL"):
            join_nulls = (None,) * (len(table.columns) - len(source_rows[0]) if source_rows else 0)
            for index, row in enumerate(source_rows):
                if index not in matched_source:
                    table.append(row + join_nulls)

        if side in ("RIGHT", "FULL"):
            source_width = len(table.columns) - (len(join_rows[0]) if join_rows else 0)
            source_nulls = (None,) * source_width
            for index, row in enumerate(join_rows):
                if index not in matched_join:
                    table.append(source_nulls + row)

    def aggregate(self, step, context):
        group_by = self.generate_tuple(step.group.values())
        aggregations = self.generate_tuple(step.aggregations)
        operands = self.generate_tuple(step.operands)

        if operands:
            operand_table = Table(self.table(step.operands).columns)

            for reader, ctx in context:
                operand_table.append(ctx.eval_tuple(operands))

            for i, (a, b) in enumerate(zip(context.table.rows, operand_table.rows)):
                context.table.rows[i] = a + b

            width = len(context.columns)
            context.add_columns(*operand_table.columns)

            operand_table = Table(
                context.columns,
                context.table.rows,
                range(width, width + len(operand_table.columns)),
            )

            context = self.context(
                {
                    None: operand_table,
                    **context.tables,
                }
            )

        context.sort(group_by)

        group = None
        start = 0
        end = 1
        length = len(context.table)
        table = self.table(list(step.group) + step.aggregations)

        def add_row():
            table.append(group + context.eval_tuple(aggregations))

        if length:
            for i in range(length):
                context.set_index(i)
                key = context.eval_tuple(group_by)
                group = key if group is None else group
                end += 1
                if key != group:
                    context.set_range(start, end - 2)
                    add_row()
                    group = key
                    start = end - 2
                if len(table.rows) >= step.limit:
                    break
                if i == length - 1:
                    context.set_range(start, end - 1)
                    add_row()
        elif step.limit > 0 and not group_by:
            context.set_range(0, 0)
            table.append(context.eval_tuple(aggregations))

        context = self.context({step.name: table, **{name: table for name in context.tables}})

        if step.projections or step.condition:
            return self.scan(step, context)
        return context

    def sort(self, step, context):
        projections = self.generate_tuple(step.projections)
        projection_columns = [p.alias_or_name for p in step.projections]
        all_columns = list(context.columns) + projection_columns
        sink = self.table(all_columns)
        for reader, ctx in context:
            sink.append(reader.row + ctx.eval_tuple(projections))

        sort_ctx = self.context(
            {
                None: sink,
                **{table: sink for table in context.tables},
            }
        )
        sort_ctx.sort(self.generate_tuple(step.key))

        if not math.isinf(step.limit):
            sort_ctx.table.rows = sort_ctx.table.rows[0 : step.limit]

        output = Table(
            projection_columns,
            rows=[r[len(context.columns) : len(all_columns)] for r in sort_ctx.table.rows],
        )
        return self.context({step.name: output})

    def set_operation(self, step, context):
        left = context.tables[step.left]
        right = context.tables[step.right]

        sink = self.table(left.columns)

        if issubclass(step.op, exp.Intersect):
            right_counts = collections.Counter(right.rows)
            seen = set()
            for row in left.rows:
                if right_counts[row] and (not step.distinct or row not in seen):
                    sink.append(row)
                    seen.add(row)
                    if not step.distinct:
                        right_counts[row] -= 1
        elif issubclass(step.op, exp.Except):
            right_counts = collections.Counter(right.rows)
            seen = set()
            for row in left.rows:
                if right_counts[row] and not step.distinct:
                    right_counts[row] -= 1
                elif not right_counts[row] and (not step.distinct or row not in seen):
                    sink.append(row)
                    seen.add(row)
        elif issubclass(step.op, exp.Union) and step.distinct:
            sink.rows = list(set(left.rows).union(set(right.rows)))
        else:
            sink.rows = left.rows + right.rows

        if not math.isinf(step.limit):
            sink.rows = sink.rows[0 : step.limit]

        return self.context({step.name: sink})


class Python(Dialect):
    class Tokenizer(tokens.Tokenizer):
        STRING_ESCAPES = ["\\"]

    Generator = PythonGenerator



from __future__ import annotations

from sqlglot import tokens
from sqlglot.dialects.dialect import (
    Dialect,
    NormalizationStrategy,
)
from sqlglot.generators.mysql import MySQLGenerator
from sqlglot.parsers.mysql import MySQLParser
from sqlglot.tokens import TokenType
from sqlglot.typing.mysql import EXPRESSION_METADATA


class MySQL(Dialect):
    PROMOTE_TO_INFERRED_DATETIME_TYPE = True

    # https://dev.mysql.com/doc/refman/8.0/en/identifiers.html
    IDENTIFIERS_CAN_START_WITH_DIGIT = True

    # We default to treating all identifiers as case-sensitive, since it matches MySQL's
    # behavior on Linux systems. For MacOS and Windows systems, one can override this
    # setting by specifying `dialect="mysql, normalization_strategy = lowercase"`.
    #
    # See also https://dev.mysql.com/doc/refman/8.2/en/identifier-case-sensitivity.html
    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_SENSITIVE

    TIME_FORMAT = "'%Y-%m-%d %T'"
    DPIPE_IS_STRING_CONCAT = False
    SUPPORTS_USER_DEFINED_TYPES = False
    SAFE_DIVISION = True
    SAFE_TO_ELIMINATE_DOUBLE_NEGATION = False
    LEAST_GREATEST_IGNORES_NULLS = False

    EXPRESSION_METADATA = EXPRESSION_METADATA.copy()

    # https://prestodb.io/docs/current/functions/datetime.html#mysql-date-functions
    TIME_MAPPING = {
        "%M": "%B",
        "%c": "%-m",
        "%e": "%-d",
        "%h": "%I",
        "%i": "%M",
        "%s": "%S",
        "%u": "%W",
        "%k": "%-H",
        "%l": "%-I",
        "%T": "%H:%M:%S",
        "%W": "%A",
    }

    VALID_INTERVAL_UNITS = {
        *Dialect.VALID_INTERVAL_UNITS,
        "SECOND_MICROSECOND",
        "MINUTE_MICROSECOND",
        "MINUTE_SECOND",
        "HOUR_MICROSECOND",
        "HOUR_SECOND",
        "HOUR_MINUTE",
        "DAY_MICROSECOND",
        "DAY_SECOND",
        "DAY_MINUTE",
        "DAY_HOUR",
        "YEAR_MONTH",
    }

    class Tokenizer(tokens.Tokenizer):
        QUOTES = ["'", '"']
        COMMENTS = ["--", "#", ("/*", "*/")]
        IDENTIFIERS = ["`"]
        STRING_ESCAPES = ["'", '"', "\\"]
        BIT_STRINGS = [("b'", "'"), ("B'", "'"), ("0b", "")]
        HEX_STRINGS = [("x'", "'"), ("X'", "'"), ("0x", "")]
        # https://dev.mysql.com/doc/refman/8.4/en/string-literals.html
        ESCAPE_FOLLOW_CHARS = ["0", "b", "n", "r", "t", "Z", "%", "_"]

        NESTED_COMMENTS = False

        KEYWORDS = {
            **tokens.Tokenizer.KEYWORDS,
            "BLOB": TokenType.BLOB,
            "CHARSET": TokenType.CHARACTER_SET,
            "DISTINCTROW": TokenType.DISTINCT,
            "EXPLAIN": TokenType.DESCRIBE,
            "FORCE": TokenType.FORCE,
            "IGNORE": TokenType.IGNORE,
            "KEY": TokenType.KEY,
            "LOCK TABLES": TokenType.COMMAND,
            "LONGBLOB": TokenType.LONGBLOB,
            "LONGTEXT": TokenType.LONGTEXT,
            "MEDIUMBLOB": TokenType.MEDIUMBLOB,
            "MEDIUMINT": TokenType.MEDIUMINT,
            "MEDIUMTEXT": TokenType.MEDIUMTEXT,
            "MEMBER OF": TokenType.MEMBER_OF,
            "MOD": TokenType.MOD,
            "SEPARATOR": TokenType.SEPARATOR,
            "SERIAL": TokenType.SERIAL,
            "SIGNED": TokenType.BIGINT,
            "SIGNED INTEGER": TokenType.BIGINT,
            "START": TokenType.BEGIN,
            "TIMESTAMP": TokenType.TIMESTAMPTZ,
            "TINYBLOB": TokenType.TINYBLOB,
            "TINYTEXT": TokenType.TINYTEXT,
            "UNLOCK TABLES": TokenType.COMMAND,
            "UNSIGNED": TokenType.UBIGINT,
            "UNSIGNED INTEGER": TokenType.UBIGINT,
            "YEAR": TokenType.YEAR,
            "_ARMSCII8": TokenType.INTRODUCER,
            "_ASCII": TokenType.INTRODUCER,
            "_BIG5": TokenType.INTRODUCER,
            "_BINARY": TokenType.INTRODUCER,
            "_CP1250": TokenType.INTRODUCER,
            "_CP1251": TokenType.INTRODUCER,
            "_CP1256": TokenType.INTRODUCER,
            "_CP1257": TokenType.INTRODUCER,
            "_CP850": TokenType.INTRODUCER,
            "_CP852": TokenType.INTRODUCER,
            "_CP866": TokenType.INTRODUCER,
            "_CP932": TokenType.INTRODUCER,
            "_DEC8": TokenType.INTRODUCER,
            "_EUCJPMS": TokenType.INTRODUCER,
            "_EUCKR": TokenType.INTRODUCER,
            "_GB18030": TokenType.INTRODUCER,
            "_GB2312": TokenType.INTRODUCER,
            "_GBK": TokenType.INTRODUCER,
            "_GEOSTD8": TokenType.INTRODUCER,
            "_GREEK": TokenType.INTRODUCER,
            "_HEBREW": TokenType.INTRODUCER,
            "_HP8": TokenType.INTRODUCER,
            "_KEYBCS2": TokenType.INTRODUCER,
            "_KOI8R": TokenType.INTRODUCER,
            "_KOI8U": TokenType.INTRODUCER,
            "_LATIN1": TokenType.INTRODUCER,
            "_LATIN2": TokenType.INTRODUCER,
            "_LATIN5": TokenType.INTRODUCER,
            "_LATIN7": TokenType.INTRODUCER,
            "_MACCE": TokenType.INTRODUCER,
            "_MACROMAN": TokenType.INTRODUCER,
            "_SJIS": TokenType.INTRODUCER,
            "_SWE7": TokenType.INTRODUCER,
            "_TIS620": TokenType.INTRODUCER,
            "_UCS2": TokenType.INTRODUCER,
            "_UJIS": TokenType.INTRODUCER,
            # https://dev.mysql.com/doc/refman/8.0/en/string-literals.html
            "_UTF8": TokenType.INTRODUCER,
            "_UTF16": TokenType.INTRODUCER,
            "_UTF16LE": TokenType.INTRODUCER,
            "_UTF32": TokenType.INTRODUCER,
            "_UTF8MB3": TokenType.INTRODUCER,
            "_UTF8MB4": TokenType.INTRODUCER,
            "@@": TokenType.SESSION_PARAMETER,
        }

        COMMANDS = {*tokens.Tokenizer.COMMANDS, TokenType.REPLACE} - {TokenType.SHOW}

    Parser = MySQLParser

    Generator = MySQLGenerator



from setuptools import setup
from setuptools_scm import get_version

version = get_version(local_scheme="no-local-version")

setup(
    extras_require={
        "dev": [
            "duckdb>=0.6",
            # sqlglot-mypy 1.20+ is the build dep for sqlglotc and only ships
            # for py3.10+; on py3.9 just use upstream mypy for type checking.
            "sqlglot-mypy >= 2.3.0.post1; python_version >= '3.10'",
            "mypy; python_version < '3.10'",
            "setuptools_scm",
            "pandas",
            "pandas-stubs",
            "python-dateutil",
            "pytz",
            "pdoc",
            "pre-commit",
            "ruff==0.15.6",
            "types-python-dateutil",
            "types-pytz",
            "typing_extensions",
            "pyperf",
        ],
        # Compiles from source on the user's machine. Requires Python 3.10+
        # because the build dep (sqlglot-mypy 1.20+) dropped 3.9; on 3.9
        # `pip install sqlglot[c]` is a no-op and you just get pure-Python sqlglot.
        "c": [f"sqlglotc=={version}; python_version >= '3.10'"],
        # Deprecated: the Rust tokenizer has been replaced by sqlglotc.
        "rs": ["sqlglotrs==0.13.0", f"sqlglotc=={version}; python_version >= '3.10'"],
    },
)



from __future__ import annotations

import typing as t

from sqlglot.optimizer.journal import Journal, record
from sqlglot.optimizer.scope import Scope, build_scope


if t.TYPE_CHECKING:
    from sqlglot._typing import E


def eliminate_ctes(expression: E, journal: Journal | None = None) -> E:
    """
    Remove unused CTEs from an expression.

    Example:
        >>> import sqlglot
        >>> sql = "WITH y AS (SELECT a FROM x) SELECT a FROM z"
        >>> expression = sqlglot.parse_one(sql)
        >>> eliminate_ctes(expression).sql()
        'SELECT a FROM z'

    Args:
        expression (sqlglot.Expr): expression to optimize
    Returns:
        sqlglot.Expr: optimized expression
    """
    root = build_scope(expression)

    if root:
        ref_count = root.ref_count()

        # Traverse the scope tree in reverse so we can remove chains of unused CTEs
        for scope in reversed(list(root.traverse())):
            if scope.is_cte:
                count = ref_count[id(scope)]
                if count <= 0:
                    cte_node = scope.expression.parent
                    if not cte_node:
                        continue
                    with_node = cte_node.parent
                    if journal is not None and with_node:
                        record(journal, with_node, "expressions")
                    cte_node.pop()

                    # Pop the entire WITH clause if this is the last CTE
                    if with_node and len(with_node.expressions) <= 0:
                        if journal is not None and with_node.parent:
                            record(journal, with_node.parent, "with_")
                        with_node.pop()

                    # Decrement the ref count for all sources this CTE selects from
                    for _, source in scope.selected_sources.values():
                        if isinstance(source, Scope):
                            ref_count[id(source)] -= 1

    return expression



from __future__ import annotations

from sqlglot.dialects.mysql import MySQL
from sqlglot.generators.doris import DorisGenerator
from sqlglot.parsers.doris import DorisParser


class Doris(MySQL):
    DATE_FORMAT = "'yyyy-MM-dd'"
    DATEINT_FORMAT = "'yyyyMMdd'"
    TIME_FORMAT = "'yyyy-MM-dd HH:mm:ss'"

    Parser = DorisParser

    Generator = DorisGenerator



import doctest
import importlib
import pkgutil
import unittest

import sqlglot


def load_tests(loader, tests, ignore):
    """
    This finds and runs all the doctests
    """

    modules = set()
    for info in pkgutil.walk_packages(sqlglot.__path__, prefix="sqlglot."):
        if info.name == "sqlglot.__main__":
            continue
        try:
            modules.add(importlib.import_module(info.name))
        except Exception:
            continue

    assert len(modules) >= 20

    for module in sorted(modules, key=lambda m: m.__name__):
        tests.addTests(doctest.DocTestSuite(module))

    return tests


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations


from sqlglot import exp, transforms
from sqlglot.dialects.dialect import (
    date_delta_sql,
    timestamptrunc_sql,
    groupconcat_sql,
)
from sqlglot.generators.spark import SparkGenerator


class DatabricksGenerator(SparkGenerator):
    TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
    COPY_PARAMS_ARE_WRAPPED = False
    COPY_PARAMS_EQ_REQUIRED = True
    JSON_PATH_SINGLE_QUOTE_ESCAPE = False
    JSON_PATH_KEY_QUOTED_FORCES_BRACKETS = True
    SAFE_JSON_PATH_KEY_RE = exp.SAFE_IDENTIFIER_RE
    QUOTE_JSON_PATH = False
    PARSE_JSON_NAME: str | None = "PARSE_JSON"

    TRANSFORMS = {
        k: v
        for k, v in {
            **SparkGenerator.TRANSFORMS,
            exp.CurrentVersion: lambda *_: "CURRENT_VERSION()",
            exp.DateAdd: date_delta_sql("DATEADD"),
            exp.DateDiff: date_delta_sql("DATEDIFF"),
            exp.DatetimeAdd: lambda self, e: self.func(
                "TIMESTAMPADD", e.unit, e.expression, e.this
            ),
            exp.DatetimeSub: lambda self, e: self.func(
                "TIMESTAMPADD",
                e.unit,
                exp.Mul(this=e.expression, expression=exp.Literal.number(-1)),
                e.this,
            ),
            exp.DatetimeTrunc: timestamptrunc_sql(),
            exp.GroupConcat: groupconcat_sql,
            exp.Select: transforms.preprocess(
                [
                    transforms.eliminate_distinct_on,
                    transforms.unnest_to_explode,
                    transforms.any_to_exists,
                ]
            ),
            exp.JSONExtract: lambda self, e: f"{self.sql(e, 'this')}:{self.sql(e, 'expression')}",
            exp.JSONPathRoot: lambda self, e: (
                "$" if isinstance(e.parent and e.parent.parent, exp.JSONExtractScalar) else ""
            ),
            exp.ToChar: lambda self, e: (
                self.cast_sql(exp.Cast(this=e.this, to=exp.DataType(this="STRING")))
                if e.args.get("is_numeric")
                else self.function_fallback_sql(e)
            ),
            exp.CurrentCatalog: lambda *_: "CURRENT_CATALOG()",
            exp.RegexpLike: None,
            exp.TryCast: None,
            exp.RegrAvgx: lambda self, e: self._regr_sql(e),
            exp.RegrSxx: lambda self, e: self._regr_sql(e),
            exp.RegrSyy: lambda self, e: self._regr_sql(e),
        }.items()
        if v is not None
    }

    TYPE_MAPPING = {
        **SparkGenerator.TYPE_MAPPING,
        exp.DType.NULL: "VOID",
    }

    def create_sql(self, expression: exp.Create) -> str:
        body = expression.expression
        if (
            body
            and not isinstance(body, exp.Return)
            and expression.kind == "FUNCTION"
            and any(p.args.get("is_table") for p in expression.find_all(exp.ReturnsProperty))
        ):
            expression.set("expression", exp.Return(this=body))
        return super().create_sql(expression)

    def columndef_sql(self, expression: exp.ColumnDef, sep: str = " ") -> str:
        constraint = expression.find(exp.GeneratedAsIdentityColumnConstraint)
        kind = expression.kind
        if (
            constraint
            and isinstance(kind, exp.DataType)
            and kind.this in exp.DataType.INTEGER_TYPES
        ):
            # only BIGINT generated identity constraints are supported
            expression.set("kind", exp.DType.BIGINT.into_expr())

        return super().columndef_sql(expression, sep)

    def timeserieskey_sql(self, expression: exp.TimeseriesKey) -> str:
        return f"{self.sql(expression, 'this')} TIMESERIES"

    def jsonpath_sql(self, expression: exp.JSONPath) -> str:
        expression.set("escape", None)
        path = super().jsonpath_sql(expression)

        if isinstance(expression.parent, exp.JSONExtractScalar):
            return f"{self.dialect.QUOTE_START}{path}{self.dialect.QUOTE_END}"

        return path

    def uniform_sql(self, expression: exp.Uniform) -> str:
        gen = expression.args.get("gen")
        seed = expression.args.get("seed")

        # From Snowflake UNIFORM(min, max, gen) as RANDOM(), RANDOM(seed), or constant value -> Extract seed
        if gen:
            seed = gen.this

        return self.func("UNIFORM", expression.this, expression.expression, seed)

    def _regr_sql(self, expression: exp.RegrAvgx | exp.RegrSxx | exp.RegrSyy) -> str:
        name = expression.sql_name()
        x = expression.expression
        if isinstance(x, exp.Distinct):
            return self.func(name, exp.Distinct(expressions=[expression.this]), *x.expressions)
        return self.func(name, expression.this, x)

    def clusterproperty_sql(self, expression):
        this = self.sql(expression, "this") or f"({self.expressions(expression, flat=True)})"
        return f"CLUSTER BY {this}"



from tests.dialects.test_dialect import Validator


class TestDAX(Validator):
    dialect = "dax"

    def test_evaluate(self):
        self.validate_transpile("EVALUATE Sales", "SELECT * FROM Sales")
        self.validate_transpile("EVALUATE 'Sales Data'", 'SELECT * FROM "Sales Data"')

    def test_filter(self):
        self.validate_transpile(
            "EVALUATE FILTER(Sales, Sales[Amount] > 100)",
            'SELECT * FROM Sales WHERE Sales."Amount" > 100',
        )
        self.validate_transpile(
            "EVALUATE FILTER('Sales Data', 'Sales Data'[Amount] > 100)",
            'SELECT * FROM "Sales Data" WHERE "Sales Data"."Amount" > 100',
        )
        self.validate_transpile(
            'EVALUATE FILTER(FILTER(Sales, Sales[Amount] > 100), Sales[Region] = "West")',
            'SELECT * FROM Sales WHERE Sales."Amount" > 100 AND Sales."Region" = \'West\'',
        )
        self.validate_transpile(
            "EVALUATE FILTER(Sales, Sales[Amount] > 100 && Sales[Qty] < 5)",
            'SELECT * FROM Sales WHERE Sales."Amount" > 100 AND Sales."Qty" < 5',
        )
        self.validate_transpile(
            "EVALUATE FILTER(Sales, Sales[Amount] > 100 || Sales[Qty] < 5)",
            'SELECT * FROM Sales WHERE Sales."Amount" > 100 OR Sales."Qty" < 5',
        )
        self.validate_transpile(
            "EVALUATE FILTER(Sales, [Total Amount] > 100)",
            'SELECT * FROM Sales WHERE "Total Amount" > 100',
        )
        self.validate_transpile(
            'EVALUATE FILTER(Sales, Sales[Note] = "He said ""hi""")',
            'SELECT * FROM Sales WHERE Sales."Note" = \'He said "hi"\'',
        )
        self.validate_transpile(
            "EVALUATE FILTER(Sales, Sales[Amount] > 100)",
            'SELECT * FROM Sales WHERE Sales."Amount" > 100',
            write_dialect="duckdb",
        )

    def test_order_by(self):
        self.validate_transpile(
            "EVALUATE Sales ORDER BY Sales[Amount] DESC, Sales[Qty]",
            'SELECT * FROM Sales ORDER BY Sales."Amount" DESC, Sales."Qty"',
        )
        self.validate_transpile(
            "EVALUATE FILTER(Sales, Sales[Amount] > 100) ORDER BY Sales[Amount]",
            'SELECT * FROM Sales WHERE Sales."Amount" > 100 ORDER BY Sales."Amount" NULLS FIRST',
            write_dialect="duckdb",
        )



from __future__ import annotations


from sqlglot import exp, tokens

from sqlglot.dialects.dialect import (
    Dialect,
    NormalizationStrategy,
)
from sqlglot.generators.duckdb import DuckDBGenerator, WS_CONTROL_CHARS_TO_DUCK  # noqa: F401
from sqlglot.parsers.duckdb import DuckDBParser
from sqlglot.tokens import TokenType
from sqlglot.typing.duckdb import EXPRESSION_METADATA


class DuckDB(Dialect):
    NULL_ORDERING = "nulls_are_last"
    SUPPORTS_USER_DEFINED_TYPES = True
    INDEX_OFFSET = 1
    CONCAT_COALESCE = True
    CONCAT_WS_COALESCE = True
    SUPPORTS_ORDER_BY_ALL = True
    SUPPORTS_LIMIT_ALL = True
    SUPPORTS_FIXED_SIZE_ARRAYS = True
    STRICT_JSON_PATH_SYNTAX = False
    NUMBERS_CAN_BE_UNDERSCORE_SEPARATED = True
    UUID_IS_STRING_TYPE = False

    # https://duckdb.org/docs/sql/introduction.html#creating-a-new-table
    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE

    DATE_PART_MAPPING = {
        **Dialect.DATE_PART_MAPPING,
        "DAYOFWEEKISO": "ISODOW",
    }

    EXPRESSION_METADATA = EXPRESSION_METADATA.copy()

    DATE_PART_MAPPING.pop("WEEKDAY")

    INVERSE_TIME_MAPPING = {
        "%e": "%-d",  # BigQuery's space-padded day (%e) -> DuckDB's no-padding day (%-d)
        "%:z": "%z",  # In DuckDB %z	can represent +/-HH:MM, +/-HHMM, or +/-HH.
        "%-z": "%z",
        "%f_zero": "%n",
        "%f_one": "%n",
        "%f_two": "%n",
        "%f_three": "%g",
        "%f_four": "%n",
        "%f_five": "%n",
        "%f_seven": "%n",
        "%f_eight": "%n",
        "%f_nine": "%n",
    }

    def to_json_path(self, path: exp.Expr | None) -> exp.Expr | None:
        if isinstance(path, exp.Literal):
            # DuckDB also supports the JSON pointer syntax, where every path starts with a `/`.
            # Additionally, it allows accessing the back of lists using the `[#-i]` syntax.
            # This check ensures we'll avoid trying to parse these as JSON paths, which can
            # either result in a noisy warning or in an invalid representation of the path.
            path_text = path.name
            if path_text.startswith("/") or "[#" in path_text:
                return path

        return super().to_json_path(path)

    class Tokenizer(tokens.Tokenizer):
        BYTE_STRINGS = [("e'", "'"), ("E'", "'")]
        BYTE_STRING_ESCAPES = ["'", "\\"]
        HEREDOC_STRINGS = ["$"]

        HEREDOC_TAG_IS_IDENTIFIER = True
        HEREDOC_STRING_ALTERNATIVE = TokenType.PARAMETER

        KEYWORDS = {
            **tokens.Tokenizer.KEYWORDS,
            "//": TokenType.DIV,
            "**": TokenType.DSTAR,
            "^@": TokenType.CARET_AT,
            "@>": TokenType.AT_GT,
            "<@": TokenType.LT_AT,
            "ATTACH": TokenType.ATTACH,
            "BINARY": TokenType.VARBINARY,
            "BITSTRING": TokenType.BIT,
            "BPCHAR": TokenType.TEXT,
            "CHAR": TokenType.TEXT,
            "DATETIME": TokenType.TIMESTAMPNTZ,
            "DETACH": TokenType.DETACH,
            "FORCE": TokenType.FORCE,
            "INSTALL": TokenType.INSTALL,
            "INT8": TokenType.BIGINT,
            "LOGICAL": TokenType.BOOLEAN,
            "MACRO": TokenType.FUNCTION,
            "ONLY": TokenType.ONLY,
            "PIVOT_WIDER": TokenType.PIVOT,
            "POSITIONAL": TokenType.POSITIONAL,
            "RESET": TokenType.COMMAND,
            "ROW": TokenType.STRUCT,
            "SIGNED": TokenType.INT,
            "STRING": TokenType.TEXT,
            "SUMMARIZE": TokenType.SUMMARIZE,
            "TIMESTAMP": TokenType.TIMESTAMPNTZ,
            "TIMESTAMP_S": TokenType.TIMESTAMP_S,
            "TIMESTAMP_MS": TokenType.TIMESTAMP_MS,
            "TIMESTAMP_NS": TokenType.TIMESTAMP_NS,
            "TIMESTAMP_US": TokenType.TIMESTAMP,
            "UBIGINT": TokenType.UBIGINT,
            "UINTEGER": TokenType.UINT,
            "USMALLINT": TokenType.USMALLINT,
            "UTINYINT": TokenType.UTINYINT,
            "VARCHAR": TokenType.TEXT,
        }
        KEYWORDS.pop("/*+")

        SINGLE_TOKENS = {
            **tokens.Tokenizer.SINGLE_TOKENS,
            "$": TokenType.PARAMETER,
        }

        VAR_SINGLE_TOKENS = {"$"}

        COMMANDS = tokens.Tokenizer.COMMANDS - {TokenType.SHOW}

    Parser = DuckDBParser

    Generator = DuckDBGenerator



from __future__ import annotations

import re
import typing as t

from sqlglot import exp
from sqlglot.dialects.dialect import (
    json_extract_segments,
    json_path_key_only_name,
    rename_func,
    bool_xor_sql,
    count_if_to_sum,
    timestamptrunc_sql,
    date_add_interval_sql,
    timestampdiff_sql,
    remove_ts_or_ds_to_date,
)
from sqlglot.expressions import DataType
from sqlglot import generator
from sqlglot.generator import unsupported_args
from sqlglot.generators.mysql import MySQLGenerator, date_add_sql


def _unicode_substitute(m: re.Match[str]) -> str:
    return chr(int(m.group(1), 16))


class SingleStoreGenerator(MySQLGenerator):
    SUPPORTS_UESCAPE = False
    SUPPORTS_ALTER_COLUMN_NULLABILITY = False
    NULL_ORDERING_SUPPORTED: bool | None = True
    MATCH_AGAINST_TABLE_PREFIX: str | None = "TABLE "
    STRUCT_DELIMITER = ("(", ")")

    UNICODE_SUBSTITUTE: t.ClassVar[t.Any] = staticmethod(_unicode_substitute)

    SUPPORTED_JSON_PATH_PARTS = {
        exp.JSONPathKey,
        exp.JSONPathRoot,
        exp.JSONPathSubscript,
    }

    TRANSFORMS = {
        **{
            k: v
            for k, v in MySQLGenerator.TRANSFORMS.items()
            if k not in (exp.JSONExtractScalar, exp.CurrentDate)
        },
        exp.TsOrDsToDate: lambda self, e: (
            self.func("TO_DATE", e.this, self.format_time(e))
            if e.args.get("format")
            else self.func("DATE", e.this)
        ),
        exp.StrToTime: lambda self, e: self.func("TO_TIMESTAMP", e.this, self.format_time(e)),
        exp.ToChar: lambda self, e: self.func("TO_CHAR", e.this, self.format_time(e)),
        exp.StrToDate: lambda self, e: self.func(
            "STR_TO_DATE",
            e.this,
            self.format_time(
                e,
                inverse_time_mapping=self.dialect.MYSQL_INVERSE_TIME_MAPPING,
                inverse_time_trie=self.dialect.MYSQL_INVERSE_TIME_TRIE,
            ),
        ),
        exp.TimeToStr: lambda self, e: self.func(
            "DATE_FORMAT",
            e.this,
            self.format_time(
                e,
                inverse_time_mapping=self.dialect.MYSQL_INVERSE_TIME_MAPPING,
                inverse_time_trie=self.dialect.MYSQL_INVERSE_TIME_TRIE,
            ),
        ),
        exp.Date: unsupported_args("zone", "expressions")(rename_func("DATE")),
        exp.Cast: unsupported_args("format", "action", "default")(
            lambda self, e: f"{self.sql(e, 'this')} :> {self.sql(e, 'to')}"
        ),
        exp.TryCast: unsupported_args("format", "action", "default")(
            lambda self, e: f"{self.sql(e, 'this')} !:> {self.sql(e, 'to')}"
        ),
        exp.CastToStrType: lambda self, e: self.sql(
            exp.cast(e.this, DataType.from_str(e.args["to"].name))
        ),
        exp.StrToUnix: unsupported_args("format")(rename_func("UNIX_TIMESTAMP")),
        exp.TimeToUnix: rename_func("UNIX_TIMESTAMP"),
        exp.TimeStrToUnix: rename_func("UNIX_TIMESTAMP"),
        exp.UnixSeconds: rename_func("UNIX_TIMESTAMP"),
        exp.UnixToStr: lambda self, e: self.func(
            "FROM_UNIXTIME",
            e.this,
            self.format_time(
                e,
                inverse_time_mapping=self.dialect.MYSQL_INVERSE_TIME_MAPPING,
                inverse_time_trie=self.dialect.MYSQL_INVERSE_TIME_TRIE,
            ),
        ),
        exp.UnixToTime: unsupported_args("scale", "zone", "hours", "minutes")(
            lambda self, e: self.func(
                "FROM_UNIXTIME",
                e.this,
                self.format_time(
                    e,
                    inverse_time_mapping=self.dialect.MYSQL_INVERSE_TIME_MAPPING,
                    inverse_time_trie=self.dialect.MYSQL_INVERSE_TIME_TRIE,
                ),
            ),
        ),
        exp.UnixToTimeStr: lambda self, e: f"FROM_UNIXTIME({self.sql(e, 'this')}) :> TEXT",
        exp.DateBin: unsupported_args("unit", "zone")(
            lambda self, e: self.func("TIME_BUCKET", e.this, e.expression, e.args.get("origin"))
        ),
        exp.TimeStrToDate: lambda self, e: self.sql(exp.cast(e.this, exp.DType.DATE)),
        exp.FromTimeZone: lambda self, e: self.func(
            "CONVERT_TZ", e.this, e.args.get("zone"), "'UTC'"
        ),
        exp.DiToDate: lambda self, e: (
            f"STR_TO_DATE({self.sql(e, 'this')}, {self.dialect.DATEINT_FORMAT})"
        ),
        exp.DateToDi: lambda self, e: (
            f"(DATE_FORMAT({self.sql(e, 'this')}, {self.dialect.DATEINT_FORMAT}) :> INT)"
        ),
        exp.TsOrDiToDi: lambda self, e: (
            f"(DATE_FORMAT({self.sql(e, 'this')}, {self.dialect.DATEINT_FORMAT}) :> INT)"
        ),
        exp.Time: unsupported_args("zone")(lambda self, e: f"{self.sql(e, 'this')} :> TIME"),
        exp.DatetimeAdd: remove_ts_or_ds_to_date(date_add_sql("ADD")),
        exp.DatetimeTrunc: unsupported_args("zone")(timestamptrunc_sql()),
        exp.DatetimeSub: date_add_interval_sql("DATE", "SUB"),
        exp.DatetimeDiff: timestampdiff_sql,
        exp.DateTrunc: unsupported_args("zone")(timestamptrunc_sql()),
        exp.DateDiff: unsupported_args("zone")(
            lambda self, e: (
                timestampdiff_sql(self, e)
                if e.unit is not None
                else self.func("DATEDIFF", e.this, e.expression)
            )
        ),
        exp.TsOrDsDiff: lambda self, e: (
            timestampdiff_sql(self, e)
            if e.unit is not None
            else self.func("DATEDIFF", e.this, e.expression)
        ),
        exp.TimestampTrunc: unsupported_args("zone")(timestamptrunc_sql()),
        exp.CurrentDatetime: lambda self, e: self.sql(
            self.dialect.CAST_TO_TIME6(
                exp.CurrentTimestamp(this=exp.Literal.number(6)), exp.DType.DATETIME
            )
        ),
        exp.JSONExtract: unsupported_args(
            "only_json_types",
            "expressions",
            "variant_extract",
            "json_query",
            "option",
            "quote",
            "on_condition",
            "requires_json",
        )(json_extract_segments("JSON_EXTRACT_JSON")),
        exp.JSONBExtract: json_extract_segments("BSON_EXTRACT_BSON"),
        exp.JSONPathKey: json_path_key_only_name,
        exp.JSONPathSubscript: lambda self, e: self.json_path_part(e.this),
        exp.JSONPathRoot: lambda *_: "",
        exp.JSONFormat: unsupported_args("options", "is_json")(rename_func("JSON_PRETTY")),
        exp.JSONArrayAgg: unsupported_args("null_handling", "return_type", "strict")(
            lambda self, e: self.func("JSON_AGG", e.this, suffix=f"{self.sql(e, 'order')})")
        ),
        exp.JSONArray: unsupported_args("null_handling", "return_type", "strict")(
            rename_func("JSON_BUILD_ARRAY")
        ),
        exp.JSONBExists: lambda self, e: self.func(
            "BSON_MATCH_ANY_EXISTS", e.this, e.args.get("path")
        ),
        exp.JSONExists: lambda self, e: (
            f"{self.sql(e.this)}::?{self.sql(e.args.get('path'))}"
            if e.args.get("from_dcolonqmark")
            else self.func("JSON_MATCH_ANY_EXISTS", e.this, e.args.get("path"))
        ),
        exp.JSONObject: unsupported_args("null_handling", "unique_keys", "return_type", "encoding")(
            rename_func("JSON_BUILD_OBJECT")
        ),
        exp.DayOfWeekIso: lambda self, e: f"(({self.func('DAYOFWEEK', e.this)} % 7) + 1)",
        exp.DayOfMonth: rename_func("DAY"),
        exp.Hll: rename_func("APPROX_COUNT_DISTINCT"),
        exp.ApproxDistinct: rename_func("APPROX_COUNT_DISTINCT"),
        exp.CountIf: count_if_to_sum,
        exp.LogicalOr: lambda self, e: f"MAX(ABS({self.sql(e, 'this')}))",
        exp.LogicalAnd: lambda self, e: f"MIN(ABS({self.sql(e, 'this')}))",
        exp.ApproxQuantile: unsupported_args("accuracy", "weight")(
            lambda self, e: self.func(
                "APPROX_PERCENTILE",
                e.this,
                e.args.get("quantile"),
                e.args.get("error_tolerance"),
            )
        ),
        exp.Variance: rename_func("VAR_SAMP"),
        exp.VariancePop: rename_func("VAR_POP"),
        exp.Xor: bool_xor_sql,
        exp.Cbrt: lambda self, e: self.sql(
            exp.Pow(this=e.this, expression=exp.Literal.number(1) / exp.Literal.number(3))
        ),
        exp.RegexpLike: lambda self, e: self.binary(e, "RLIKE"),
        exp.Repeat: lambda self, e: self.func(
            "LPAD",
            exp.Literal.string(""),
            exp.Mul(this=self.func("LENGTH", e.this), expression=e.args.get("times")),
            e.this,
        ),
        exp.IsAscii: lambda self, e: f"({self.sql(e, 'this')} RLIKE '^[\x00-\x7f]*$')",
        exp.MD5Digest: lambda self, e: self.func("UNHEX", self.func("MD5", e.this)),
        exp.Contains: rename_func("INSTR"),
        exp.RegexpExtractAll: unsupported_args("position", "occurrence", "group")(
            lambda self, e: self.func(
                "REGEXP_MATCH",
                e.this,
                e.expression,
                e.args.get("parameters"),
            )
        ),
        exp.RegexpExtract: unsupported_args("group")(
            lambda self, e: self.func(
                "REGEXP_SUBSTR",
                e.this,
                e.expression,
                e.args.get("position"),
                e.args.get("occurrence"),
                e.args.get("parameters"),
            )
        ),
        exp.StartsWith: lambda self, e: self.func(
            "REGEXP_INSTR", e.this, self.func("CONCAT", exp.Literal.string("^"), e.expression)
        ),
        exp.FromBase: lambda self, e: self.func(
            "CONV", e.this, e.expression, exp.Literal.number(10)
        ),
        exp.RegexpILike: lambda self, e: self.binary(
            exp.RegexpLike(
                this=exp.Lower(this=e.this),
                expression=exp.Lower(this=e.expression),
            ),
            "RLIKE",
        ),
        exp.Stuff: lambda self, e: self.func(
            "CONCAT",
            self.func("SUBSTRING", e.this, exp.Literal.number(1), e.args.get("start") - 1),
            e.expression,
            self.func("SUBSTRING", e.this, e.args.get("start") + e.args.get("length")),
        ),
        exp.National: lambda self, e: self.national_sql(e, prefix=""),
        exp.Reduce: unsupported_args("finish")(
            lambda self, e: self.func("REDUCE", e.args.get("initial"), e.this, e.args.get("merge"))
        ),
        exp.MatchAgainst: unsupported_args("modifier")(
            lambda self, e: generator.Generator.matchagainst_sql(self, e)
        ),
        exp.Show: unsupported_args(
            "history",
            "terse",
            "offset",
            "starts_with",
            "limit",
            "from_",
            "scope",
            "scope_kind",
            "mutex",
            "query",
            "channel",
            "log",
            "types",
            "privileges",
        )(lambda self, e: MySQLGenerator.show_sql(self, e)),
        exp.Describe: unsupported_args(
            "style",
            "kind",
            "expressions",
            "partition",
            "format",
        )(lambda self, e: generator.Generator.describe_sql(self, e)),
    }

    UNSUPPORTED_TYPES = {
        exp.DType.ARRAY,
        exp.DType.AGGREGATEFUNCTION,
        exp.DType.SIMPLEAGGREGATEFUNCTION,
        exp.DType.BIGSERIAL,
        exp.DType.BPCHAR,
        exp.DType.DATEMULTIRANGE,
        exp.DType.DATERANGE,
        exp.DType.DYNAMIC,
        exp.DType.HLLSKETCH,
        exp.DType.HSTORE,
        exp.DType.IMAGE,
        exp.DType.INET,
        exp.DType.INT128,
        exp.DType.INT256,
        exp.DType.INT4MULTIRANGE,
        exp.DType.INT4RANGE,
        exp.DType.INT8MULTIRANGE,
        exp.DType.INT8RANGE,
        exp.DType.INTERVAL,
        exp.DType.IPADDRESS,
        exp.DType.IPPREFIX,
        exp.DType.IPV4,
        exp.DType.IPV6,
        exp.DType.LIST,
        exp.DType.MAP,
        exp.DType.LOWCARDINALITY,
        exp.DType.MONEY,
        exp.DType.MULTILINESTRING,
        exp.DType.NAME,
        exp.DType.NESTED,
        exp.DType.NOTHING,
        exp.DType.NULL,
        exp.DType.NUMMULTIRANGE,
        exp.DType.NUMRANGE,
        exp.DType.OBJECT,
        exp.DType.RANGE,
        exp.DType.ROWVERSION,
        exp.DType.SERIAL,
        exp.DType.SMALLSERIAL,
        exp.DType.SMALLMONEY,
        exp.DType.SUPER,
        exp.DType.TIMETZ,
        exp.DType.TIMESTAMPNTZ,
        exp.DType.TIMESTAMPLTZ,
        exp.DType.TIMESTAMPTZ,
        exp.DType.TIMESTAMP_NS,
        exp.DType.TSMULTIRANGE,
        exp.DType.TSRANGE,
        exp.DType.TSTZMULTIRANGE,
        exp.DType.TSTZRANGE,
        exp.DType.UINT128,
        exp.DType.UINT256,
        exp.DType.UNION,
        exp.DType.UNKNOWN,
        exp.DType.USERDEFINED,
        exp.DType.UUID,
        exp.DType.VARIANT,
        exp.DType.XML,
        exp.DType.TDIGEST,
    }

    TYPE_MAPPING = {
        **MySQLGenerator.TYPE_MAPPING,
        exp.DType.BIGDECIMAL: "DECIMAL",
        exp.DType.BIT: "BOOLEAN",
        exp.DType.DATE32: "DATE",
        exp.DType.DATETIME64: "DATETIME",
        exp.DType.DECIMAL32: "DECIMAL",
        exp.DType.DECIMAL64: "DECIMAL",
        exp.DType.DECIMAL128: "DECIMAL",
        exp.DType.DECIMAL256: "DECIMAL",
        exp.DType.ENUM8: "ENUM",
        exp.DType.ENUM16: "ENUM",
        exp.DType.FIXEDSTRING: "TEXT",
        exp.DType.GEOMETRY: "GEOGRAPHY",
        exp.DType.POINT: "GEOGRAPHYPOINT",
        exp.DType.RING: "GEOGRAPHY",
        exp.DType.LINESTRING: "GEOGRAPHY",
        exp.DType.POLYGON: "GEOGRAPHY",
        exp.DType.MULTIPOLYGON: "GEOGRAPHY",
        exp.DType.STRUCT: "RECORD",
        exp.DType.JSONB: "BSON",
        exp.DType.TIMESTAMP: "TIMESTAMP",
        exp.DType.TIMESTAMP_S: "TIMESTAMP",
        exp.DType.TIMESTAMP_MS: "TIMESTAMP",
    }

    TYPE_PARAM_SETTINGS = {
        **MySQLGenerator.TYPE_PARAM_SETTINGS,
        exp.DType.TIMESTAMP_MS: ((6,), ()),
    }

    # https://docs.singlestore.com/cloud/reference/sql-reference/restricted-keywords/list-of-restricted-keywords/
    RESERVED_KEYWORDS = {
        "abs",
        "absolute",
        "access",
        "account",
        "acos",
        "action",
        "add",
        "adddate",
        "addtime",
        "admin",
        "aes_decrypt",
        "aes_encrypt",
        "after",
        "against",
        "aggregate",
        "aggregates",
        "aggregator",
        "aggregator_id",
        "aggregator_plan_hash",
        "aggregators",
        "algorithm",
        "all",
        "also",
        "alter",
        "always",
        "analyse",
        "analyze",
        "and",
        "anti_join",
        "any",
        "any_value",
        "approx_count_distinct",
        "approx_count_distinct_accumulate",
        "approx_count_distinct_combine",
        "approx_count_distinct_estimate",
        "approx_geography_intersects",
        "approx_percentile",
        "arghistory",
        "arrange",
        "arrangement",
        "array",
        "as",
        "asc",
        "ascii",
        "asensitive",
        "asin",
        "asm",
        "assertion",
        "assignment",
        "ast",
        "asymmetric",
        "async",
        "at",
        "atan",
        "atan2",
        "attach",
        "attribute",
        "authorization",
        "auto",
        "auto_increment",
        "auto_reprovision",
        "autostats",
        "autostats_cardinality_mode",
        "autostats_enabled",
        "autostats_histogram_mode",
        "autostats_sampling",
        "availability",
        "avg",
        "avg_row_length",
        "avro",
        "azure",
        "background",
        "_background_threads_for_cleanup",
        "backup",
        "backup_history",
        "backup_id",
        "backward",
        "batch",
        "batches",
        "batch_interval",
        "_batch_size_limit",
        "before",
        "begin",
        "between",
        "bigint",
        "bin",
        "binary",
        "_binary",
        "bit",
        "bit_and",
        "bit_count",
        "bit_or",
        "bit_xor",
        "blob",
        "bool",
        "boolean",
        "bootstrap",
        "both",
        "_bt",
        "btree",
        "bucket_count",
        "by",
        "byte",
        "byte_length",
        "cache",
        "call",
        "call_for_pipeline",
        "called",
        "capture",
        "cascade",
        "cascaded",
        "case",
        "cast",
        "catalog",
        "ceil",
        "ceiling",
        "chain",
        "change",
        "char",
        "character",
        "characteristics",
        "character_length",
        "char_length",
        "charset",
        "check",
        "checkpoint",
        "_check_can_connect",
        "_check_consistency",
        "checksum",
        "_checksum",
        "class",
        "clear",
        "client",
        "client_found_rows",
        "close",
        "cluster",
        "clustered",
        "cnf",
        "coalesce",
        "coercibility",
        "collate",
        "collation",
        "collect",
        "column",
        "columnar",
        "columns",
        "columnstore",
        "columnstore_segment_rows",
        "comment",
        "comments",
        "commit",
        "committed",
        "_commit_log_tail",
        "committed",
        "compact",
        "compile",
        "compressed",
        "compression",
        "concat",
        "concat_ws",
        "concurrent",
        "concurrently",
        "condition",
        "configuration",
        "connection",
        "connection_id",
        "connections",
        "config",
        "constraint",
        "constraints",
        "content",
        "continue",
        "_continue_replay",
        "conv",
        "conversion",
        "convert",
        "convert_tz",
        "copy",
        "_core",
        "cos",
        "cost",
        "cot",
        "count",
        "create",
        "credentials",
        "cross",
        "cube",
        "csv",
        "cume_dist",
        "curdate",
        "current",
        "current_catalog",
        "current_date",
        "current_role",
        "current_schema",
        "current_security_groups",
        "current_security_roles",
        "current_time",
        "current_timestamp",
        "current_user",
        "cursor",
        "curtime",
        "cycle",
        "data",
        "database",
        "databases",
        "date",
        "date_add",
        "datediff",
        "date_format",
        "date_sub",
        "date_trunc",
        "datetime",
        "day",
        "day_hour",
        "day_microsecond",
        "day_minute",
        "dayname",
        "dayofmonth",
        "dayofweek",
        "dayofyear",
        "day_second",
        "deallocate",
        "dec",
        "decimal",
        "declare",
        "decode",
        "default",
        "defaults",
        "deferrable",
        "deferred",
        "defined",
        "definer",
        "degrees",
        "delayed",
        "delay_key_write",
        "delete",
        "delimiter",
        "delimiters",
        "dense_rank",
        "desc",
        "describe",
        "detach",
        "deterministic",
        "dictionary",
        "differential",
        "directory",
        "disable",
        "discard",
        "_disconnect",
        "disk",
        "distinct",
        "distinctrow",
        "distributed_joins",
        "div",
        "do",
        "document",
        "domain",
        "dot_product",
        "double",
        "drop",
        "_drop_profile",
        "dual",
        "dump",
        "duplicate",
        "dynamic",
        "earliest",
        "each",
        "echo",
        "election",
        "else",
        "elseif",
        "elt",
        "enable",
        "enclosed",
        "encoding",
        "encrypted",
        "end",
        "engine",
        "engines",
        "enum",
        "errors",
        "escape",
        "escaped",
        "estimate",
        "euclidean_distance",
        "event",
        "events",
        "except",
        "exclude",
        "excluding",
        "exclusive",
        "execute",
        "exists",
        "exit",
        "exp",
        "explain",
        "extended",
        "extension",
        "external",
        "external_host",
        "external_port",
        "extract",
        "extractor",
        "extractors",
        "extra_join",
        "_failover",
        "failed_login_attempts",
        "failure",
        "false",
        "family",
        "fault",
        "fetch",
        "field",
        "fields",
        "file",
        "files",
        "fill",
        "first",
        "first_value",
        "fix_alter",
        "fixed",
        "float",
        "float4",
        "float8",
        "floor",
        "flush",
        "following",
        "for",
        "force",
        "force_compiled_mode",
        "force_interpreter_mode",
        "foreground",
        "foreign",
        "format",
        "forward",
        "found_rows",
        "freeze",
        "from",
        "from_base64",
        "from_days",
        "from_unixtime",
        "fs",
        "_fsync",
        "full",
        "fulltext",
        "function",
        "functions",
        "gc",
        "gcs",
        "get_format",
        "_gc",
        "_gcx",
        "generate",
        "geography",
        "geography_area",
        "geography_contains",
        "geography_distance",
        "geography_intersects",
        "geography_latitude",
        "geography_length",
        "geography_longitude",
        "geographypoint",
        "geography_point",
        "geography_within_distance",
        "geometry",
        "geometry_area",
        "geometry_contains",
        "geometry_distance",
        "geometry_filter",
        "geometry_intersects",
        "geometry_length",
        "geometrypoint",
        "geometry_point",
        "geometry_within_distance",
        "geometry_x",
        "geometry_y",
        "global",
        "_global_version_timestamp",
        "grant",
        "granted",
        "grants",
        "greatest",
        "group",
        "grouping",
        "groups",
        "group_concat",
        "gzip",
        "handle",
        "handler",
        "hard_cpu_limit_percentage",
        "hash",
        "has_temp_tables",
        "having",
        "hdfs",
        "header",
        "heartbeat_no_logging",
        "hex",
        "highlight",
        "high_priority",
        "hold",
        "holding",
        "host",
        "hosts",
        "hour",
        "hour_microsecond",
        "hour_minute",
        "hour_second",
        "identified",
        "identity",
        "if",
        "ifnull",
        "ignore",
        "ilike",
        "immediate",
        "immutable",
        "implicit",
        "import",
        "in",
        "including",
        "increment",
        "incremental",
        "index",
        "indexes",
        "inet_aton",
        "inet_ntoa",
        "inet6_aton",
        "inet6_ntoa",
        "infile",
        "inherit",
        "inherits",
        "_init_profile",
        "init",
        "initcap",
        "initialize",
        "initially",
        "inject",
        "inline",
        "inner",
        "inout",
        "input",
        "insensitive",
        "insert",
        "insert_method",
        "instance",
        "instead",
        "instr",
        "int",
        "int1",
        "int2",
        "int3",
        "int4",
        "int8",
        "integer",
        "_internal_dynamic_typecast",
        "interpreter_mode",
        "intersect",
        "interval",
        "into",
        "invoker",
        "is",
        "isnull",
        "isolation",
        "iterate",
        "join",
        "json",
        "json_agg",
        "json_array_contains_double",
        "json_array_contains_json",
        "json_array_contains_string",
        "json_array_push_double",
        "json_array_push_json",
        "json_array_push_string",
        "json_delete_key",
        "json_extract_double",
        "json_extract_json",
        "json_extract_string",
        "json_extract_bigint",
        "json_get_type",
        "json_length",
        "json_set_double",
        "json_set_json",
        "json_set_string",
        "json_splice_double",
        "json_splice_json",
        "json_splice_string",
        "kafka",
        "key",
        "key_block_size",
        "keys",
        "kill",
        "killall",
        "label",
        "lag",
        "language",
        "large",
        "last",
        "last_day",
        "last_insert_id",
        "last_value",
        "lateral",
        "latest",
        "lc_collate",
        "lc_ctype",
        "lcase",
        "lead",
        "leading",
        "leaf",
        "leakproof",
        "least",
        "leave",
        "leaves",
        "left",
        "length",
        "level",
        "license",
        "like",
        "limit",
        "lines",
        "listen",
        "llvm",
        "ln",
        "load",
        "loaddata_where",
        "_load",
        "local",
        "localtime",
        "localtimestamp",
        "locate",
        "location",
        "lock",
        "log",
        "log10",
        "log2",
        "long",
        "longblob",
        "longtext",
        "loop",
        "lower",
        "low_priority",
        "lpad",
        "_ls",
        "ltrim",
        "lz4",
        "management",
        "_management_thread",
        "mapping",
        "master",
        "match",
        "materialized",
        "max",
        "maxvalue",
        "max_concurrency",
        "max_errors",
        "max_partitions_per_batch",
        "max_queue_depth",
        "max_retries_per_batch_partition",
        "max_rows",
        "mbc",
        "md5",
        "mpl",
        "median",
        "mediumblob",
        "mediumint",
        "mediumtext",
        "member",
        "memory",
        "memory_percentage",
        "_memsql_table_id_lookup",
        "memsql",
        "memsql_deserialize",
        "memsql_imitating_kafka",
        "memsql_serialize",
        "merge",
        "metadata",
        "microsecond",
        "middleint",
        "min",
        "min_rows",
        "minus",
        "minute",
        "minute_microsecond",
        "minute_second",
        "minvalue",
        "mod",
        "mode",
        "model",
        "modifies",
        "modify",
        "month",
        "monthname",
        "months_between",
        "move",
        "mpl",
        "names",
        "named",
        "namespace",
        "national",
        "natural",
        "nchar",
        "next",
        "no",
        "node",
        "none",
        "no_query_rewrite",
        "noparam",
        "not",
        "nothing",
        "notify",
        "now",
        "nowait",
        "no_write_to_binlog",
        "no_query_rewrite",
        "norely",
        "nth_value",
        "ntile",
        "null",
        "nullcols",
        "nullif",
        "nulls",
        "numeric",
        "nvarchar",
        "object",
        "octet_length",
        "of",
        "off",
        "offline",
        "offset",
        "offsets",
        "oids",
        "on",
        "online",
        "only",
        "open",
        "operator",
        "optimization",
        "optimize",
        "optimizer",
        "optimizer_state",
        "option",
        "options",
        "optionally",
        "or",
        "order",
        "ordered_serialize",
        "orphan",
        "out",
        "out_of_order",
        "outer",
        "outfile",
        "over",
        "overlaps",
        "overlay",
        "owned",
        "owner",
        "pack_keys",
        "paired",
        "parser",
        "parquet",
        "partial",
        "partition",
        "partition_id",
        "partitioning",
        "partitions",
        "passing",
        "password",
        "password_lock_time",
        "parser",
        "pause",
        "_pause_replay",
        "percent_rank",
        "percentile_cont",
        "percentile_disc",
        "periodic",
        "persisted",
        "pi",
        "pipeline",
        "pipelines",
        "pivot",
        "placing",
        "plan",
        "plans",
        "plancache",
        "plugins",
        "pool",
        "pools",
        "port",
        "position",
        "pow",
        "power",
        "preceding",
        "precision",
        "prepare",
        "prepared",
        "preserve",
        "primary",
        "prior",
        "privileges",
        "procedural",
        "procedure",
        "procedures",
        "process",
        "processlist",
        "profile",
        "profiles",
        "program",
        "promote",
        "proxy",
        "purge",
        "quarter",
        "queries",
        "query",
        "query_timeout",
        "queue",
        "quote",
        "radians",
        "rand",
        "range",
        "rank",
        "read",
        "_read",
        "reads",
        "real",
        "reassign",
        "rebalance",
        "recheck",
        "record",
        "recursive",
        "redundancy",
        "redundant",
        "ref",
        "reference",
        "references",
        "refresh",
        "regexp",
        "reindex",
        "relative",
        "release",
        "reload",
        "rely",
        "remote",
        "remove",
        "rename",
        "repair",
        "_repair_table",
        "repeat",
        "repeatable",
        "_repl",
        "_reprovisioning",
        "replace",
        "replica",
        "replicate",
        "replicating",
        "replication",
        "durability",
        "require",
        "resource",
        "resource_pool",
        "reset",
        "restart",
        "restore",
        "restrict",
        "result",
        "_resurrect",
        "retry",
        "return",
        "returning",
        "returns",
        "reverse",
        "revoke",
        "rg_pool",
        "right",
        "right_anti_join",
        "right_semi_join",
        "right_straight_join",
        "rlike",
        "role",
        "roles",
        "rollback",
        "rollup",
        "round",
        "routine",
        "row",
        "row_count",
        "row_format",
        "row_number",
        "rows",
        "rowstore",
        "rule",
        "rpad",
        "_rpc",
        "rtrim",
        "running",
        "s3",
        "safe",
        "save",
        "savepoint",
        "scalar",
        "schema",
        "schemas",
        "schema_binding",
        "scroll",
        "search",
        "second",
        "second_microsecond",
        "sec_to_time",
        "security",
        "select",
        "semi_join",
        "_send_threads",
        "sensitive",
        "separator",
        "sequence",
        "sequences",
        "serial",
        "serializable",
        "series",
        "service_user",
        "server",
        "session",
        "session_user",
        "set",
        "setof",
        "security_lists_intersect",
        "sha",
        "sha1",
        "sha2",
        "shard",
        "sharded",
        "sharded_id",
        "share",
        "show",
        "shutdown",
        "sigmoid",
        "sign",
        "signal",
        "similar",
        "simple",
        "site",
        "signed",
        "sin",
        "skip",
        "skipped_batches",
        "sleep",
        "_sleep",
        "smallint",
        "snapshot",
        "_snapshot",
        "_snapshots",
        "soft_cpu_limit_percentage",
        "some",
        "soname",
        "sparse",
        "spatial",
        "spatial_check_index",
        "specific",
        "split",
        "sql",
        "sql_big_result",
        "sql_buffer_result",
        "sql_cache",
        "sql_calc_found_rows",
        "sqlexception",
        "sql_mode",
        "sql_no_cache",
        "sql_no_logging",
        "sql_small_result",
        "sqlstate",
        "sqlwarning",
        "sqrt",
        "ssl",
        "stable",
        "standalone",
        "start",
        "starting",
        "state",
        "statement",
        "statistics",
        "stats",
        "status",
        "std",
        "stddev",
        "stddev_pop",
        "stddev_samp",
        "stdin",
        "stdout",
        "stop",
        "storage",
        "str_to_date",
        "straight_join",
        "strict",
        "string",
        "strip",
        "subdate",
        "substr",
        "substring",
        "substring_index",
        "success",
        "sum",
        "super",
        "symmetric",
        "sync_snapshot",
        "sync",
        "_sync",
        "_sync2",
        "_sync_partitions",
        "_sync_snapshot",
        "synchronize",
        "sysid",
        "system",
        "table",
        "table_checksum",
        "tables",
        "tablespace",
        "tags",
        "tan",
        "target_size",
        "task",
        "temp",
        "template",
        "temporary",
        "temptable",
        "_term_bump",
        "terminate",
        "terminated",
        "test",
        "text",
        "then",
        "time",
        "timediff",
        "time_bucket",
        "time_format",
        "timeout",
        "timestamp",
        "timestampadd",
        "timestampdiff",
        "timezone",
        "time_to_sec",
        "tinyblob",
        "tinyint",
        "tinytext",
        "to",
        "to_base64",
        "to_char",
        "to_date",
        "to_days",
        "to_json",
        "to_number",
        "to_seconds",
        "to_timestamp",
        "tracelogs",
        "traditional",
        "trailing",
        "transform",
        "transaction",
        "_transactions_experimental",
        "treat",
        "trigger",
        "triggers",
        "trim",
        "true",
        "trunc",
        "truncate",
        "trusted",
        "two_phase",
        "_twopcid",
        "type",
        "types",
        "ucase",
        "unbounded",
        "uncommitted",
        "undefined",
        "undo",
        "unencrypted",
        "unenforced",
        "unhex",
        "unhold",
        "unicode",
        "union",
        "unique",
        "_unittest",
        "unix_timestamp",
        "unknown",
        "unlisten",
        "_unload",
        "unlock",
        "unlogged",
        "unpivot",
        "unsigned",
        "until",
        "update",
        "upgrade",
        "upper",
        "usage",
        "use",
        "user",
        "users",
        "using",
        "utc_date",
        "utc_time",
        "utc_timestamp",
        "_utf8",
        "vacuum",
        "valid",
        "validate",
        "validator",
        "value",
        "values",
        "varbinary",
        "varchar",
        "varcharacter",
        "variables",
        "variadic",
        "variance",
        "var_pop",
        "var_samp",
        "varying",
        "vector_sub",
        "verbose",
        "version",
        "view",
        "void",
        "volatile",
        "voting",
        "wait",
        "_wake",
        "warnings",
        "week",
        "weekday",
        "weekofyear",
        "when",
        "where",
        "while",
        "whitespace",
        "window",
        "with",
        "without",
        "within",
        "_wm_heartbeat",
        "work",
        "workload",
        "wrapper",
        "write",
        "xact_id",
        "xor",
        "year",
        "year_month",
        "yes",
        "zerofill",
        "zone",
    }

    def jsonextractscalar_sql(self, expression: exp.JSONExtractScalar) -> str:
        json_type = expression.args.get("json_type")
        func_name = "JSON_EXTRACT_JSON" if json_type is None else f"JSON_EXTRACT_{json_type}"
        return json_extract_segments(func_name)(self, expression)

    def jsonbextractscalar_sql(self, expression: exp.JSONBExtractScalar) -> str:
        json_type = expression.args.get("json_type")
        func_name = "BSON_EXTRACT_BSON" if json_type is None else f"BSON_EXTRACT_{json_type}"
        return json_extract_segments(func_name)(self, expression)

    def jsonextractarray_sql(self, expression: exp.JSONExtractArray) -> str:
        self.unsupported("Arrays are not supported in SingleStore")
        return self.function_fallback_sql(expression)

    def jsonvalue_sql(self, expression: exp.JSONValue) -> str:
        if expression.args.get("on_condition"):
            self.unsupported("JSON_VALUE does not support on_condition")
        res: exp.Expr = exp.JSONExtractScalar(
            this=expression.this,
            expression=expression.args.get("path"),
            json_type="STRING",
        )

        returning = expression.args.get("returning")
        if returning is not None:
            res = exp.Cast(this=res, to=returning)

        return self.sql(res)

    def all_sql(self, expression: exp.All) -> str:
        self.unsupported("ALL subquery predicate is not supported in SingleStore")
        return super().all_sql(expression)

    def jsonarraycontains_sql(self, expression: exp.JSONArrayContains) -> str:
        json_type = expression.text("json_type").upper()

        if json_type:
            return self.func(
                f"JSON_ARRAY_CONTAINS_{json_type}", expression.expression, expression.this
            )

        return self.func(
            "JSON_ARRAY_CONTAINS_JSON",
            expression.expression,
            self.func("TO_JSON", expression.this),
        )

    def datatype_sql(self, expression: exp.DataType) -> str:
        for arg_name in ("kind", "values"):
            if expression.args.get(arg_name):
                self.unsupported(f"DATATYPE does not support {arg_name}")
        if expression.args.get("nested") and not expression.is_type(exp.DType.STRUCT):
            self.unsupported(
                f"Argument 'nested' is not supported for representation of '{expression.this.value}' in SingleStore"
            )

        if expression.is_type(exp.DType.VARBINARY) and not expression.expressions:
            # `VARBINARY` must always have a size - if it doesn't, we always generate `BLOB`
            return "BLOB"
        if expression.is_type(
            exp.DType.DECIMAL32,
            exp.DType.DECIMAL64,
            exp.DType.DECIMAL128,
            exp.DType.DECIMAL256,
        ):
            scale = self.expressions(expression, flat=True)

            if expression.is_type(exp.DType.DECIMAL32):
                precision = "9"
            elif expression.is_type(exp.DType.DECIMAL64):
                precision = "18"
            elif expression.is_type(exp.DType.DECIMAL128):
                precision = "38"
            else:
                # 65 is a maximum precision supported in SingleStore
                precision = "65"
            if scale is not None:
                return f"DECIMAL({precision}, {scale[0]})"
            else:
                return f"DECIMAL({precision})"
        if expression.is_type(exp.DType.VECTOR):
            expressions = expression.expressions
            if len(expressions) == 2:
                type_name = self.sql(expressions[0])
                if type_name in self.dialect.INVERSE_VECTOR_TYPE_ALIASES:
                    type_name = self.dialect.INVERSE_VECTOR_TYPE_ALIASES[type_name]

                return f"VECTOR({self.sql(expressions[1])}, {type_name})"

        return super().datatype_sql(expression)

    def collate_sql(self, expression: exp.Collate) -> str:
        # SingleStore does not support setting a collation for column in the SELECT query,
        # so we cast column to a LONGTEXT type with specific collation
        return self.binary(expression, ":> LONGTEXT COLLATE")

    def currentdate_sql(self, expression: exp.CurrentDate) -> str:
        timezone = expression.this
        if timezone:
            if isinstance(timezone, exp.Literal) and timezone.name.lower() == "utc":
                return self.func("UTC_DATE")
            self.unsupported("CurrentDate with timezone is not supported in SingleStore")

        return self.func("CURRENT_DATE")

    def currenttime_sql(self, expression: exp.CurrentTime) -> str:
        arg = expression.this
        if arg:
            if isinstance(arg, exp.Literal) and arg.name.lower() == "utc":
                return self.func("UTC_TIME")
            if isinstance(arg, exp.Literal) and arg.is_number:
                return self.func("CURRENT_TIME", arg)
            self.unsupported("CurrentTime with timezone is not supported in SingleStore")

        return self.func("CURRENT_TIME")

    def currenttimestamp_sql(self, expression: exp.CurrentTimestamp) -> str:
        arg = expression.this
        if arg:
            if isinstance(arg, exp.Literal) and arg.name.lower() == "utc":
                return self.func("UTC_TIMESTAMP")
            if isinstance(arg, exp.Literal) and arg.is_number:
                return self.func("CURRENT_TIMESTAMP", arg)
            self.unsupported("CurrentTimestamp with timezone is not supported in SingleStore")

        return self.func("CURRENT_TIMESTAMP")

    def standardhash_sql(self, expression: exp.StandardHash) -> str:
        hash_function = expression.expression
        if hash_function is None:
            return self.func("SHA", expression.this)
        if isinstance(hash_function, exp.Literal):
            if hash_function.name.lower() == "sha":
                return self.func("SHA", expression.this)
            if hash_function.name.lower() == "md5":
                return self.func("MD5", expression.this)

            self.unsupported(f"{hash_function.this} hash method is not supported in SingleStore")
            return self.func("SHA", expression.this)

        self.unsupported("STANDARD_HASH function is not supported in SingleStore")
        return self.func("SHA", expression.this)

    def truncatetable_sql(self, expression: exp.TruncateTable) -> str:
        for arg_name in ("is_database", "exists", "cluster", "identity", "option", "partition"):
            if expression.args.get(arg_name):
                self.unsupported(f"TRUNCATE TABLE does not support {arg_name}")
        statements = []
        for table in expression.expressions:
            statements.append(f"TRUNCATE {self.sql(table)}")

        return "; ".join(statements)

    def renamecolumn_sql(self, expression: exp.RenameColumn) -> str:
        if expression.args.get("exists"):
            self.unsupported("RENAME COLUMN does not support exists")
        old_column = self.sql(expression, "this")
        new_column = self.sql(expression, "to")
        return f"CHANGE {old_column} {new_column}"

    def altercolumn_sql(self, expression: exp.AlterColumn) -> str:
        for arg_name in ("drop", "comment", "visible", "using"):
            if expression.args.get(arg_name):
                self.unsupported(f"ALTER COLUMN does not support {arg_name}")
        alter = super().altercolumn_sql(expression)

        collate = self.sql(expression, "collate")
        collate = f" COLLATE {collate}" if collate else ""
        return f"{alter}{collate}"

    def computedcolumnconstraint_sql(self, expression: exp.ComputedColumnConstraint) -> str:
        this = self.sql(expression, "this")
        not_null = " NOT NULL" if expression.args.get("not_null") else ""
        type = self.sql(expression, "data_type") or "AUTO"
        return f"AS {this} PERSISTED {type}{not_null}"
